import os
import asyncio
import json
import re
from collections.abc import Callable, Generator
from pathlib import Path
from typing import Literal
from uuid import uuid4

import httpx
from a2a.client import A2ACardResolver, A2AClient
from a2a.types import (
    AgentCard,
    Message,
    MessageSendParams,
    Part,
    Role,
    SendStreamingMessageRequest,
    SendStreamingMessageSuccessResponse,
    TaskStatusUpdateEvent,
    TextPart,
)
from google import genai
from jinja2 import Template
from dotenv import load_dotenv

load_dotenv()

GOOGLE_API_KEY = os.environ.get("GOOGLE_API_KEY")

dir_path = Path(__file__).parent

with Path(dir_path / "tpl" / "decide.jinja").open("r") as f:
    decide_template = Template(f.read())

with Path(dir_path / "tpl" / "agents.jinja").open("r") as f:
    agents_template = Template(f.read())

with Path(dir_path / "tpl" / "agent_answer.jinja").open("r") as f:
    agent_answer_template = Template(f.read())


def stream_llm(prompt: str) -> Generator[str]:
    """Stream LLM response.

    Args:
        prompt (str): The prompt to send to the LLM.

    Returns:
        Generator[str, None, None]: A generator of the LLM response.
    !TODO: use openai / local inference engine (vLLM/SGLang/TensorRT-LLM wrapped by achatbot)
    """
    client = genai.Client(vertexai=False, api_key=GOOGLE_API_KEY)
    for chunk in client.models.generate_content_stream(
        model="gemini-2.5-flash-lite",
        contents=prompt,
    ):
        yield chunk.text


class Agent:
    """Agent for interacting with the Google Gemini LLM in different modes."""

    def __init__(
        self,
        mode: Literal["complete", "stream"] = "stream",
        token_stream_callback: Callable[[str], None] | None = None,
        agent_urls: list[str] | None = None,
    ):
        self.mode = mode
        self.token_stream_callback = token_stream_callback
        self.agent_urls = agent_urls
        self.agents_registry: dict[str, AgentCard] = {}

    async def get_agents(self) -> tuple[dict[str, AgentCard], str]:
        """Retrieve agent cards from all agent URLs and render the agent prompt.

        Returns:
            tuple[dict[str, AgentCard], str]: A dictionary mapping agent names to AgentCard objects, and the rendered agent prompt string.
        """
        async with httpx.AsyncClient() as httpx_client:
            card_resolvers = [A2ACardResolver(httpx_client, url) for url in self.agent_urls]
            agent_cards = await asyncio.gather(
                *[card_resolver.get_agent_card() for card_resolver in card_resolvers]
            )
            agents_registry = {agent_card.name: agent_card for agent_card in agent_cards}
            agent_prompt = agents_template.render(agent_cards=agent_cards)
            return agents_registry, agent_prompt

    def call_llm(self, prompt: str) -> str:
        """Call the LLM with the given prompt and return the response as a string or generator.

        Args:
            prompt (str): The prompt to send to the LLM.

        Returns:
            str or Generator[str]: The LLM response as a string or generator, depending on mode.
        """
        if self.mode == "complete":
            return stream_llm(prompt)

        result = ""
        for chunk in stream_llm(prompt):
            result += chunk
        return result

    async def decide(
        self,
        question: str,
        agents_prompt: str,
        called_agents: list[dict] | None = None,
    ) -> Generator[str, None]:
        """Decide which agent(s) to use to answer the question.

        Args:
            question (str): The question to answer.
            agents_prompt (str): The prompt describing available agents.
            called_agents (list[dict] | None): Previously called agents and their answers.

        Returns:
            Generator[str, None]: The LLM's response as a generator of strings.
        """
        if called_agents:
            call_agent_prompt = agent_answer_template.render(called_agents=called_agents)
        else:
            call_agent_prompt = ""
        prompt = decide_template.render(
            question=question,
            agent_prompt=agents_prompt,
            call_agent_prompt=call_agent_prompt,
        )
        return self.call_llm(prompt)

    def extract_agents(self, response: str) -> list[dict]:
        """Extract the agents from the response.

        Args:
            response (str): The response from the LLM.
        """
        pattern = r"```json\n(.*?)\n```"
        match = re.search(pattern, response, re.DOTALL)
        if match:
            return json.loads(match.group(1))
        return []

    async def send_message_to_an_agent(self, agent_card: AgentCard, message: str):
        """Send a message to a specific agent and yield the streaming response.

        Args:
            agent_card (AgentCard): The agent to send the message to.
            message (str): The message to send.

        Yields:
            str: The streaming response from the agent.
        """
        async with httpx.AsyncClient(timeout=60) as httpx_client:
            client = A2AClient(httpx_client, agent_card=agent_card)
            message = MessageSendParams(
                message=Message(
                    role=Role.user,
                    parts=[Part(TextPart(text=message))],
                    message_id=uuid4().hex,
                    # task_id=uuid4().hex,
                )
            )

            streaming_request = SendStreamingMessageRequest(id=str(uuid4().hex), params=message)
            async for chunk in client.send_message_streaming(streaming_request):
                if isinstance(chunk.root, SendStreamingMessageSuccessResponse) and isinstance(
                    chunk.root.result, TaskStatusUpdateEvent
                ):
                    message = chunk.root.result.status.message
                    if message:
                        yield message.parts[0].root.text

    async def stream(self, question: str):
        """Stream the process of answering a question, possibly involving multiple agents.

        Args:
            question (str): The question to answer.

        Yields:
            str: Streaming output, including agent responses and intermediate steps.
        """
        agent_answers: list[dict] = []
        for _ in range(3):
            agents_registry, agent_prompt = await self.get_agents()
            response = ""
            for chunk in await self.decide(question, agent_prompt, agent_answers):
                response += chunk
                if self.token_stream_callback:
                    self.token_stream_callback(chunk)
                yield chunk

            agents = self.extract_agents(response)
            if agents:
                for agent in agents:
                    agent_response = ""
                    agent_card = agents_registry[agent["name"]]
                    yield f'<Agent name="{agent["name"]}">\n'
                    async for chunk in self.send_message_to_an_agent(agent_card, agent["prompt"]):
                        # print("agent chunk--->", chunk)
                        agent_response += chunk
                        if self.token_stream_callback:
                            self.token_stream_callback(chunk)
                        yield chunk
                    yield "</Agent>\n"
                    match = re.search(r"<Answer>(.*?)</Answer>", agent_response, re.DOTALL)
                    answer = match.group(1).strip() if match else agent_response
                    # print("answer-->", answer)
                    agent_answers.append(
                        {
                            "name": agent["name"],
                            "prompt": agent["prompt"],
                            "answer": answer,
                        }
                    )
            else:
                return


"""
python -m a2a_server.cmd.client.agent
"""
if __name__ == "__main__":
    import asyncio
    import colorama

    async def main():
        """Main function to run the Agent client."""
        agent = Agent(
            mode="stream",
            token_stream_callback=None,
            agent_urls=["http://localhost:6666/"],
        )

        async for chunk in agent.stream("What is achatbot?"):
            if chunk.startswith('<Agent name="'):
                print(colorama.Fore.CYAN + chunk, end="", flush=True)
                pass
            elif chunk.startswith("</Agent>"):
                print(colorama.Fore.RESET + chunk, end="", flush=True)
                pass
            else:
                print(chunk, end="", flush=True)
                pass

    asyncio.run(main())



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch

src_ckpt = "/checkpoint/wnhsu/w2v/archived/hubert_base_ls960_it2.pt"
ref_ckpt = "/checkpoint/wnhsu/w2v/hubert_icassp_oss_v3/iter2_km100-400k-grp-L6/oss.km500_p0_1_s334.pmw1_0.puw0_0.grpnorm.ml10.mp0_8.untie.mxsz250000.ufreq1.maxtok1400000.MU100k.s1337.ngpu32/checkpoint_last.pt"
new_ckpt = "/checkpoint/wnhsu/w2v/archived/hubert_base_ls960_it2_updated.pt"


def update_state(state):
    state["model"]["label_embs_concat"] = state["model"].pop("label_embs")
    state["args"].task = "hubert_pretraining"
    state["args"].labels = f"['{state['args'].labels}']"
    return state


src_state = torch.load(src_ckpt)
src_state = update_state(src_state)
torch.save(src_state, new_ckpt)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Wrapper around various loggers and progress bars (e.g., tqdm).
"""

import atexit
import json
import logging
import os
import sys
from collections import OrderedDict
from contextlib import contextmanager
from numbers import Number
from typing import Optional

import torch

from .meters import AverageMeter, StopwatchMeter, TimeMeter


logger = logging.getLogger(__name__)


def progress_bar(
    iterator,
    log_format: Optional[str] = None,
    log_interval: int = 100,
    log_file: Optional[str] = None,
    epoch: Optional[int] = None,
    prefix: Optional[str] = None,
    tensorboard_logdir: Optional[str] = None,
    default_log_format: str = "tqdm",
    wandb_project: Optional[str] = None,
    wandb_run_name: Optional[str] = None,
    azureml_logging: Optional[bool] = False,
):
    if log_format is None:
        log_format = default_log_format
    if log_file is not None:
        handler = logging.FileHandler(filename=log_file)
        logger.addHandler(handler)

    if log_format == "tqdm" and not sys.stderr.isatty():
        log_format = "simple"

    if log_format == "json":
        bar = JsonProgressBar(iterator, epoch, prefix, log_interval)
    elif log_format == "none":
        bar = NoopProgressBar(iterator, epoch, prefix)
    elif log_format == "simple":
        bar = SimpleProgressBar(iterator, epoch, prefix, log_interval)
    elif log_format == "tqdm":
        bar = TqdmProgressBar(iterator, epoch, prefix)
    else:
        raise ValueError("Unknown log format: {}".format(log_format))

    if tensorboard_logdir:
        try:
            # [FB only] custom wrapper for TensorBoard
            import palaas  # noqa
            from .fb_tbmf_wrapper import FbTbmfWrapper

            bar = FbTbmfWrapper(bar, log_interval)
        except ImportError:
            bar = TensorboardProgressBarWrapper(bar, tensorboard_logdir)

    if wandb_project:
        bar = WandBProgressBarWrapper(bar, wandb_project, run_name=wandb_run_name)

    if azureml_logging:
        bar = AzureMLProgressBarWrapper(bar)

    return bar


def build_progress_bar(
    args,
    iterator,
    epoch: Optional[int] = None,
    prefix: Optional[str] = None,
    default: str = "tqdm",
    no_progress_bar: str = "none",
):
    """Legacy wrapper that takes an argparse.Namespace."""
    if getattr(args, "no_progress_bar", False):
        default = no_progress_bar
    if getattr(args, "distributed_rank", 0) == 0:
        tensorboard_logdir = getattr(args, "tensorboard_logdir", None)
    else:
        tensorboard_logdir = None
    return progress_bar(
        iterator,
        log_format=args.log_format,
        log_interval=args.log_interval,
        epoch=epoch,
        prefix=prefix,
        tensorboard_logdir=tensorboard_logdir,
        default_log_format=default,
    )


def format_stat(stat):
    if isinstance(stat, Number):
        stat = "{:g}".format(stat)
    elif isinstance(stat, AverageMeter):
        stat = "{:.3f}".format(stat.avg)
    elif isinstance(stat, TimeMeter):
        stat = "{:g}".format(round(stat.avg))
    elif isinstance(stat, StopwatchMeter):
        stat = "{:g}".format(round(stat.sum))
    elif torch.is_tensor(stat):
        stat = stat.tolist()
    return stat


class BaseProgressBar(object):
    """Abstract class for progress bars."""

    def __init__(self, iterable, epoch=None, prefix=None):
        self.iterable = iterable
        self.n = getattr(iterable, "n", 0)
        self.epoch = epoch
        self.prefix = ""
        if epoch is not None:
            self.prefix += "epoch {:03d}".format(epoch)
        if prefix is not None:
            self.prefix += (" | " if self.prefix != "" else "") + prefix

    def __len__(self):
        return len(self.iterable)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False

    def __iter__(self):
        raise NotImplementedError

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        raise NotImplementedError

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        raise NotImplementedError

    def update_config(self, config):
        """Log latest configuration."""
        pass

    def _str_commas(self, stats):
        return ", ".join(key + "=" + stats[key].strip() for key in stats.keys())

    def _str_pipes(self, stats):
        return " | ".join(key + " " + stats[key].strip() for key in stats.keys())

    def _format_stats(self, stats):
        postfix = OrderedDict(stats)
        # Preprocess stats according to datatype
        for key in postfix.keys():
            postfix[key] = str(format_stat(postfix[key]))
        return postfix


@contextmanager
def rename_logger(logger, new_name):
    old_name = logger.name
    if new_name is not None:
        logger.name = new_name
    yield logger
    logger.name = old_name


class JsonProgressBar(BaseProgressBar):
    """Log output in JSON format."""

    def __init__(self, iterable, epoch=None, prefix=None, log_interval=1000):
        super().__init__(iterable, epoch, prefix)
        self.log_interval = log_interval
        self.i = 0
        self.size = len(self.iterable)

    """
    def __iter__(self):
        self.size = len(self.iterable)
        for i, obj in enumerate(self.iterable, start=self.n):
            self.i = i
            yield obj
    """
    def __iter__(self):
        self.size = len(self.iterable)
        self.iterable = iter(self.iterable)
        return self

    def __next__(self):
        obj = next(self.iterable)
        self.i += 1
        return obj

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        step = step or self.i or 0
        if step > 0 and self.log_interval is not None and step % self.log_interval == 0:
            update = (
                self.epoch - 1 + (self.i + 1) / float(self.size)
                if self.epoch is not None
                else None
            )
            stats = self._format_stats(stats, epoch=self.epoch, update=update)
            with rename_logger(logger, tag):
                logger.info(json.dumps(stats))

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        self.stats = stats
        if tag is not None:
            self.stats = OrderedDict(
                [(tag + "_" + k, v) for k, v in self.stats.items()]
            )
        stats = self._format_stats(self.stats, epoch=self.epoch)
        with rename_logger(logger, tag):
            logger.info(json.dumps(stats))

    def _format_stats(self, stats, epoch=None, update=None):
        postfix = OrderedDict()
        if epoch is not None:
            postfix["epoch"] = epoch
        if update is not None:
            postfix["update"] = round(update, 3)
        # Preprocess stats according to datatype
        for key in stats.keys():
            postfix[key] = format_stat(stats[key])
        return postfix


class NoopProgressBar(BaseProgressBar):
    """No logging."""

    def __init__(self, iterable, epoch=None, prefix=None):
        super().__init__(iterable, epoch, prefix)

    def __iter__(self):
        for obj in self.iterable:
            yield obj

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        pass

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        pass


class SimpleProgressBar(BaseProgressBar):
    """A minimal logger for non-TTY environments."""

    def __init__(self, iterable, epoch=None, prefix=None, log_interval=1000):
        super().__init__(iterable, epoch, prefix)
        self.log_interval = log_interval
        self.i = None
        self.size = None

    def __iter__(self):
        self.size = len(self.iterable)
        for i, obj in enumerate(self.iterable, start=self.n):
            self.i = i
            yield obj

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        step = step or self.i or 0
        if step > 0 and self.log_interval is not None and step % self.log_interval == 0:
            stats = self._format_stats(stats)
            postfix = self._str_commas(stats)
            with rename_logger(logger, tag):
                logger.info(
                    "{}:  {:5d} / {:d} {}".format(
                        self.prefix, self.i + 1, self.size, postfix
                    )
                )

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        postfix = self._str_pipes(self._format_stats(stats))
        with rename_logger(logger, tag):
            logger.info("{} | {}".format(self.prefix, postfix))


class TqdmProgressBar(BaseProgressBar):
    """Log to tqdm."""

    def __init__(self, iterable, epoch=None, prefix=None):
        super().__init__(iterable, epoch, prefix)
        from tqdm import tqdm

        self.tqdm = tqdm(
            iterable,
            self.prefix,
            leave=False,
            disable=(logger.getEffectiveLevel() > logging.INFO),
        )

    def __iter__(self):
        return iter(self.tqdm)

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats according to log_interval."""
        self.tqdm.set_postfix(self._format_stats(stats), refresh=False)

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        postfix = self._str_pipes(self._format_stats(stats))
        with rename_logger(logger, tag):
            logger.info("{} | {}".format(self.prefix, postfix))


try:
    _tensorboard_writers = {}
    from torch.utils.tensorboard import SummaryWriter
except ImportError:
    try:
        from tensorboardX import SummaryWriter
    except ImportError:
        SummaryWriter = None


def _close_writers():
    for w in _tensorboard_writers.values():
        w.close()


atexit.register(_close_writers)


class TensorboardProgressBarWrapper(BaseProgressBar):
    """Log to tensorboard."""

    def __init__(self, wrapped_bar, tensorboard_logdir):
        self.wrapped_bar = wrapped_bar
        self.tensorboard_logdir = tensorboard_logdir

        if SummaryWriter is None:
            logger.warning(
                "tensorboard not found, please install with: pip install tensorboard"
            )

    def _writer(self, key):
        if SummaryWriter is None:
            return None
        _writers = _tensorboard_writers
        if key not in _writers:
            _writers[key] = SummaryWriter(os.path.join(self.tensorboard_logdir, key))
            _writers[key].add_text("sys.argv", " ".join(sys.argv))
        return _writers[key]

    def __iter__(self):
        return iter(self.wrapped_bar)

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats to tensorboard."""
        self._log_to_tensorboard(stats, tag, step)
        self.wrapped_bar.log(stats, tag=tag, step=step)

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        self._log_to_tensorboard(stats, tag, step)
        self.wrapped_bar.print(stats, tag=tag, step=step)

    def update_config(self, config):
        """Log latest configuration."""
        # TODO add hparams to Tensorboard
        self.wrapped_bar.update_config(config)

    def _log_to_tensorboard(self, stats, tag=None, step=None):
        writer = self._writer(tag or "")
        if writer is None:
            return
        if step is None:
            step = stats["num_updates"]
        for key in stats.keys() - {"num_updates"}:
            if isinstance(stats[key], AverageMeter):
                writer.add_scalar(key, stats[key].val, step)
            elif isinstance(stats[key], Number):
                writer.add_scalar(key, stats[key], step)
            elif torch.is_tensor(stats[key]) and stats[key].numel() == 1:
                writer.add_scalar(key, stats[key].item(), step)
        writer.flush()


try:
    import wandb
except ImportError:
    wandb = None


class WandBProgressBarWrapper(BaseProgressBar):
    """Log to Weights & Biases."""

    def __init__(self, wrapped_bar, wandb_project, run_name=None):
        self.wrapped_bar = wrapped_bar
        if wandb is None:
            logger.warning("wandb not found, pip install wandb")
            return

        # reinit=False to ensure if wandb.init() is called multiple times
        # within one process it still references the same run
        wandb.init(project=wandb_project, reinit=False, name=run_name)

    def __iter__(self):
        return iter(self.wrapped_bar)

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats to tensorboard."""
        self._log_to_wandb(stats, tag, step)
        self.wrapped_bar.log(stats, tag=tag, step=step)

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats."""
        self._log_to_wandb(stats, tag, step)
        self.wrapped_bar.print(stats, tag=tag, step=step)

    def update_config(self, config):
        """Log latest configuration."""
        if wandb is not None:
            wandb.config.update(config)
        self.wrapped_bar.update_config(config)

    def _log_to_wandb(self, stats, tag=None, step=None):
        if wandb is None:
            return
        if step is None:
            step = stats["num_updates"]

        prefix = "" if tag is None else tag + "/"

        for key in stats.keys() - {"num_updates"}:
            if isinstance(stats[key], AverageMeter):
                wandb.log({prefix + key: stats[key].val}, step=step)
            elif isinstance(stats[key], Number):
                wandb.log({prefix + key: stats[key]}, step=step)


try:
    from azureml.core import Run
except ImportError:
    Run = None


class AzureMLProgressBarWrapper(BaseProgressBar):
    """Log to Azure ML"""

    def __init__(self, wrapped_bar):
        self.wrapped_bar = wrapped_bar
        if Run is None:
            logger.warning("azureml.core not found, pip install azureml-core")
            return
        self.run = Run.get_context()

    def __exit__(self, *exc):
        if Run is not None:
            self.run.complete()
        return False

    def __iter__(self):
        return iter(self.wrapped_bar)

    def log(self, stats, tag=None, step=None):
        """Log intermediate stats to AzureML"""
        self._log_to_azureml(stats, tag, step)
        self.wrapped_bar.log(stats, tag=tag, step=step)

    def print(self, stats, tag=None, step=None):
        """Print end-of-epoch stats"""
        self._log_to_azureml(stats, tag, step)
        self.wrapped_bar.print(stats, tag=tag, step=step)

    def update_config(self, config):
        """Log latest configuration."""
        self.wrapped_bar.update_config(config)

    def _log_to_azureml(self, stats, tag=None, step=None):
        if Run is None:
            return
        if step is None:
            step = stats["num_updates"]

        prefix = "" if tag is None else tag + "/"

        for key in stats.keys() - {"num_updates"}:
            name = prefix + key
            if isinstance(stats[key], AverageMeter):
                self.run.log_row(name=name, **{"step": step, key: stats[key].val})
            elif isinstance(stats[key], Number):
                self.run.log_row(name=name, **{"step": step, key: stats[key]})



import os
import modal

app = modal.App("decoupled")

tritonserver_image = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/tritonserver
    # default install python3.12.3, in /usr/bin/python3.12.3
    modal.Image.from_registry(
        "nvcr.io/nvidia/tritonserver:25.02-trtllm-python-py3",
        add_python="3.12",  # modal install /usr/local/bin/python3.12.1
    )
    .apt_install(
        "tree",
        "git",
        "git-lfs",
        "cmake",
        "rapidjson-dev",
        "libarchive-dev",
        "zlib1g-dev",
    )
    .run_commands(
        # "update-alternatives --install /usr/local/bin/python python3 /usr/local/bin/python3.12 1",
        # "update-alternatives --install /usr/local/bin/python python3 /usr/bin/python3.12 2",
        # "python --version",
    )
    .run_commands(
        "cmake --version",  # cmake>=3.17
        "python --version",
        "mkdir -p /stub",
        "git clone https://github.com/triton-inference-server/python_backend -b r25.02",
        "cd python_backend && mkdir build",
        "cd /python_backend/build && cmake -DPYBIND11_FINDPYTHON=ON -DPYTHON_EXECUTABLE=$(which python) -DTRITON_ENABLE_GPU=ON -DTRITON_BACKEND_REPO_TAG=r25.02 -DTRITON_COMMON_REPO_TAG=r25.02 -DTRITON_CORE_REPO_TAG=r25.02 -DCMAKE_INSTALL_PREFIX:PATH=/stub/ .. && make triton-python-backend-stub",
    )
    .run_commands(
        "mkdir -p models/bls_decoupled_sync/1",
        "mkdir -p models/bls_decoupled_async/1",
        "mkdir -p /models/repeat_int32/1",
        "mkdir -p /models/square_int32/1",
        "cp /python_backend/examples/bls_decoupled/sync_model.py models/bls_decoupled_sync/1/model.py",
        "cp /python_backend/examples/bls_decoupled/sync_config.pbtxt models/bls_decoupled_sync/config.pbtxt",
        "cp /python_backend/build/triton_python_backend_stub /models/bls_decoupled_sync",
        "cp /python_backend/examples/bls_decoupled/async_model.py models/bls_decoupled_async/1/model.py",
        "cp /python_backend/examples/bls_decoupled/async_config.pbtxt models/bls_decoupled_async/config.pbtxt",
        "cp /python_backend/build/triton_python_backend_stub /models/bls_decoupled_async",
        "cp /python_backend/examples/decoupled/repeat_model.py /models/repeat_int32/1/model.py",
        "cp /python_backend/examples/decoupled/repeat_config.pbtxt /models/repeat_int32/config.pbtxt",
        "cp /python_backend/build/triton_python_backend_stub /models/repeat_int32",
        "cp /python_backend/examples/decoupled/square_model.py /models/square_int32/1/model.py",
        "cp /python_backend/examples/decoupled/square_config.pbtxt /models/square_int32/config.pbtxt",
        "cp /python_backend/build/triton_python_backend_stub /models/square_int32",
    )
    .pip_install(
        "numpy",
    )
    .env(
        {
            # "PYTHONPATH": os.getenv("IMAGE_PYTHONPATH", ""),
            # "CUDA_VISIBLE_DEVICES": os.getenv("CUDA_VISIBLE_DEVICES", "0"),
            "PYTHONIOENCODING": "utf-8",
            # "APP_NAME": os.getenv("APP_NAME", "tts-spark"),
        }
    )
)


# ⭐️ https://github.com/triton-inference-server/python_backend/blob/main/README.md
@app.function(
    # gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    image=tritonserver_image,
    volumes={},
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=100,
)
def run():
    import subprocess

    print("--" * 20)
    cmd = f"ldd /opt/tritonserver/backends/python/triton_python_backend_stub"
    print(cmd)
    subprocess.run(cmd, shell=True, check=True)

    print("--" * 20)
    cmd = f"ldd /python_backend/build/triton_python_backend_stub"
    print(cmd)
    subprocess.run(cmd, shell=True, check=True)

    cmd = f"tree /models"
    print(cmd)
    subprocess.run(cmd, shell=True, check=True)

    with modal.forward(8001, unencrypted=True) as tunnel:
        print(
            f"use tunnel.tcp_socket = {tunnel.tcp_socket[0]}:{tunnel.tcp_socket[1]} to connect tritonserver with tcp(grpc)"
        )
        cmd = f"tritonserver --model-repository /models --log-verbose 0 --log-info True"
        print(cmd)
        subprocess.run(cmd, shell=True, check=True)


# ⭐️ https://github.com/triton-inference-server/python_backend/blob/main/README.md
@app.function(
    # gpu=os.getenv("IMAGE_GPU", "L4"),
    cpu=2.0,
    image=tritonserver_image,
    volumes={},
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=100,
)
@modal.web_server(port=8000, startup_timeout=5 * 60)
def serve():
    import subprocess

    print("--" * 20)
    cmd = f"ldd /opt/tritonserver/backends/python/triton_python_backend_stub"
    print(cmd)
    subprocess.run(cmd, shell=True, check=True)

    print("--" * 20)
    cmd = f"ldd /python_backend/build/triton_python_backend_stub"
    print(cmd)
    subprocess.run(cmd, shell=True, check=True)

    cmd = f"tree /models"
    print(cmd)
    subprocess.run(cmd, shell=True, check=True)

    cmd = f"tritonserver --model-repository /models"
    print(cmd)
    subprocess.Popen(cmd, env=os.environ, shell=True)


"""
# run http tritonserver
modal serve src/llm/trtllm/decoupled/tritonserver.py 

# curl http server is ready
curl -vv -X GET "https://weedge--decoupled-serve-dev.modal.run/v2/health/ready" -H  "accept: application/json"

# run grpc tritonserver by tcp tunnel and http server
modal run src/llm/trtllm/decoupled/tritonserver.py 
"""


@app.local_entrypoint()
def main():
    run.remote()



from .base import VideoGenProcessor


def get_video_gen_processor(tag, **kwargs) -> VideoGenProcessor:
    if tag == "ComfyUIAPIVideoGenProcessor":
        from .comfyui_video_gen_processor import ComfyUIAPIVideoGenProcessor

        return ComfyUIAPIVideoGenProcessor(**kwargs)

    return None



from typing import Optional, Union

import lightning.pytorch as pl
import torch
from lightning import LightningModule, Trainer
from lightning.pytorch.callbacks import Callback
from torch import Tensor, nn
from torch.utils._foreach_utils import (
    _group_tensors_by_device_and_dtype,
    _has_foreach_support,
)


@torch.no_grad()
def grad_norm(
    parameters: Union[Tensor, list[Tensor]],
    norm_type: float = 2.0,
) -> float:
    """
    Returns the norm of the gradients of the given parameters.

    Args:
        parameters (Iterable[Tensor] or Tensor): an iterable of Tensors or a
            single Tensor that will have gradients normalized
        norm_type (float): type of the used p-norm.

    Returns:
        Total norm of the parameter gradients (viewed as a single vector).
    """  # noqa: E501

    if isinstance(parameters, Tensor):
        parameters = [parameters]

    grads = [p.grad for p in parameters if p.grad is not None]
    if len(grads) == 0:
        return None

    first_device = grads[0].device
    grouped_grads: dict[
        tuple[torch.device, torch.dtype], list[list[Tensor]]
    ] = _group_tensors_by_device_and_dtype(
        [[g.detach() for g in grads]]
    )  # type: ignore[assignment]

    norms = []
    for (device, _), ([grads], _) in grouped_grads.items():
        if _has_foreach_support(grads, device=device):
            norms.extend(torch._foreach_norm(grads, norm_type))
        else:
            norms.extend([torch.norm(g, norm_type) for g in grads])

    return torch.norm(torch.stack([norm.to(first_device) for norm in norms]), norm_type)


class GradNormMonitor(Callback):
    """
    Callback that computes the gradient norm of the model parameters.
    """

    def __init__(
        self,
        norm_type: float = 2.0,
        logging_interval: str = "step",
        sub_module: Optional[Union[str, list[str]]] = None,
    ) -> None:
        """
        Args:
            norm_type (float): type of the used p-norm.
            logging_interval (str): "step" or "epoch".
        """
        super().__init__()

        self.norm_type = norm_type
        self.logging_interval = logging_interval
        self.sub_module = sub_module

    def on_after_backward(self, trainer: Trainer, model: LightningModule) -> None:
        """
        Computes the gradient norm of the model parameters and logs it to the logger.

        Args:
            trainer (Trainer): The trainer object
            model (LightningModule): The current lightningModule
        """

        lightning_model = model

        if self.sub_module is None:
            return self.log_sub_module_grad_norm(lightning_model, model, "")

        sub_modules = self.sub_module
        if isinstance(sub_modules, str):
            sub_modules = [sub_modules]

        for sub_module in sub_modules:
            self.log_sub_module_grad_norm(
                lightning_model, getattr(model, sub_module), f"/{sub_module}"
            )

    def log_sub_module_grad_norm(
        self, lightning_model: LightningModule, model: nn.Module, path: str
    ) -> None:
        grad_norm_val = grad_norm(model.parameters(), self.norm_type)
        if grad_norm_val is None:
            return

        on_step = self.logging_interval == "step"
        lightning_model.log(
            f"train{path}/grad_norm",
            grad_norm_val,
            on_step=on_step,
            on_epoch=not on_step,
        )



import logging
from typing import Any

from apipeline.frames.control_frames import EndFrame

from src.types.frames import TransportMessageFrame
from src.services.webrtc_peer_connection import SmallWebRTCConnection
from src.cmd.bots.base import AIBot
from src.transports.small_webrtc import SmallWebRTCTransport


class SmallWebrtcAIBot(AIBot):
    def __init__(self, webrtc_connection: SmallWebRTCConnection | None = None, **args) -> None:
        super().__init__(**args)
        self._webrtc_connection = webrtc_connection

    def set_webrtc_connection(self, webrtc_connection: SmallWebRTCConnection):
        self._webrtc_connection = webrtc_connection

    def register_event(self, transport: SmallWebRTCTransport):
        transport.add_event_handler(
            "on_client_connected",
            self.on_client_connected,
        )
        transport.add_event_handler(
            "on_client_disconnected",
            self.on_client_disconnected,
        )
        transport.add_event_handler(
            "on_app_message",
            self.on_app_message,
        )

    async def on_client_connected(
        self,
        transport: SmallWebRTCTransport,
        connection: SmallWebRTCConnection,
    ):
        logging.info(f"on_client_connected {connection.pc_id=} {connection.connectionState=}")
        self.session.set_client_id(connection.pc_id)
        message = TransportMessageFrame(
            message={"type": "meta", "protocol": "small-webrtc", "version": "0.0.1"},
            urgent=True,
        )
        await transport.output_processor().send_message(message)

    async def on_client_disconnected(
        self,
        transport: SmallWebRTCTransport,
        connection: SmallWebRTCConnection,
    ):
        logging.info(f"on_client_disconnected callee id:{connection.pc_id}")
        if self.task is not None:
            await self.task.queue_frame(EndFrame())

    async def on_app_message(
        self,
        transport: SmallWebRTCTransport,
        connection: SmallWebRTCConnection,
        message: Any,
    ):
        logging.info(f"on_app_message received message: {message}")



from typing import Union
from aiortc import MediaStreamTrack
from av.frame import Frame
from av.packet import Packet


class BaseTrack(MediaStreamTrack):
    """
    Custom media stream Base track
    - from remote peer track as in track for bot peer to process
    - set video/audio kind to process frame
    - default recv is echo back
    """

    def __init__(self, in_track: MediaStreamTrack = None) -> None:
        assert in_track is not None
        super().__init__()

        self.in_track = in_track
        self.kind = in_track.kind

    def set_kind(self, kind: str) -> None:
        self.kind = kind

    def set_in_track(self, in_track: MediaStreamTrack) -> None:
        self.in_track = in_track

    async def recv(self) -> Union[Frame, Packet]:
        """
        echo
        Receive the next :class:`~av.audio.frame.AudioFrame`,
        :class:`~av.video.frame.VideoFrame` or :class:`~av.packet.Packet`
        """
        assert self.in_track is not None
        frame = await self.in_track.recv()
        # print(frame)
        return frame



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""isort:skip_file"""

import importlib
import os

from fairseq import registry
from fairseq.criterions.fairseq_criterion import (  # noqa
    FairseqCriterion,
    LegacyFairseqCriterion,
)
from omegaconf import DictConfig


(
    build_criterion_,
    register_criterion,
    CRITERION_REGISTRY,
    CRITERION_DATACLASS_REGISTRY,
) = registry.setup_registry(
    "--criterion", base_class=FairseqCriterion, default="cross_entropy"
)


def build_criterion(cfg: DictConfig, task):
    return build_criterion_(cfg, task)


# automatically import any Python files in the criterions/ directory
for file in sorted(os.listdir(os.path.dirname(__file__))):
    if file.endswith(".py") and not file.startswith("_"):
        file_name = file[: file.find(".py")]
        importlib.import_module("fairseq.criterions." + file_name)



import time
import logging
from typing import AsyncGenerator, cast
import asyncio

from apipeline.processors.frame_processor import FrameDirection
from apipeline.frames import Frame, StartFrame, EndFrame, CancelFrame
import numpy as np

from achatbot.modules.avatar.lam_audio2expression import LAMAudio2ExpressionAvatar
from achatbot.types.avatar import SpeechAudio, AudioSlice, AvatarStatus
from achatbot.processors.avatar.base import AvatarProcessorBase, SegmentedAvatarProcessor
from achatbot.processors.avatar.help.speech_audio_slicer import SpeechAudioSlicer
from achatbot.types.frames import AudioRawFrame, AnimationAudioRawFrame
from achatbot.common.utils.audio_utils import bytes2NpArrayWith16
from achatbot.types.frames.control_frames import (
    UserStartedSpeakingFrame,
    UserStoppedSpeakingFrame,
    TTSStartedFrame,
    TTSStoppedFrame,
)


# class LAMAudio2ExpressionAvatarProcessor(AvatarProcessorBase):
class LAMAudio2ExpressionAvatarProcessor(SegmentedAvatarProcessor):
    def __init__(self, avatar: LAMAudio2ExpressionAvatar, **kwargs):
        self._avatar = avatar
        self.input_audio_slice_duration = float(kwargs.get("input_audio_slice_duration", "1"))
        super().__init__(sample_rate=self._avatar.args.speaker_audio_sample_rate, **kwargs)

        # running
        self._session_running = False
        self._session_start_time = 0

        # running task
        self._audio2expression_task: asyncio.Task = None
        # running asyncio task queue
        self._audio_slice_queue: asyncio.Queue = None

        # audio input slice
        self._speech_audio_slicer: SpeechAudioSlicer = None

        # don't to load avatar, outside to load, processor just to run with session
        # self._avatar.load()

    async def start(self, frame: StartFrame):
        await super().start(frame)
        self._session_running = True
        self._init()
        self._start_tasks()
        self._session_start_time = time.time()

    def _init(self):
        self._speech_audio_slicer = SpeechAudioSlicer(
            self._avatar.args.speaker_audio_sample_rate,  # input
            self._avatar.args.avatar_audio_sample_rate,  # output for avatar input audio sample rate
            self.input_audio_slice_duration,
        )

    def _start_tasks(self):
        self._audio_slice_queue = asyncio.Queue()
        self._audio2expression_task = self.get_event_loop().create_task(
            self._audio2expression_loop()
        )

    async def stop(self, frame: EndFrame):
        await super().stop(frame)
        await self._stop()

    async def cancel(self, frame: CancelFrame):
        await super().cancel(frame)
        await self._stop()

    async def _stop(self):
        logging.info(
            f"stop avatar processor, totol session time {time.time() - self._session_start_time:.3f}",
        )
        self._session_running = False
        if self._audio2expression_task is not None:
            self._audio2expression_task.cancel()
            await self._audio2expression_task
            self._audio2expression_task = None
        logging.info("avatar processor stopped")

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, UserStartedSpeakingFrame):
            await self.queue_frame(
                AnimationAudioRawFrame(
                    audio=b"",
                    avatar_status=str(AvatarStatus.LISTENING),
                )
            )
        elif isinstance(frame, UserStoppedSpeakingFrame):
            await self.queue_frame(
                AnimationAudioRawFrame(
                    audio=b"",
                    avatar_status=str(AvatarStatus.THINKING),
                )
            )
        if isinstance(frame, TTSStartedFrame):
            await self.queue_frame(
                AnimationAudioRawFrame(
                    audio=b"",
                    avatar_status=str(AvatarStatus.RESPONDING),
                )
            )
        elif isinstance(frame, TTSStoppedFrame):
            await self.queue_frame(
                AnimationAudioRawFrame(
                    audio=b"",
                    avatar_status=str(AvatarStatus.LISTENING),
                )
            )

    async def run_avatar(self, frame: AudioRawFrame) -> AsyncGenerator[Frame, None]:
        await self._add_audio(
            SpeechAudio(
                end_of_speech=True,
                speech_id="",
                sample_rate=frame.sample_rate,
                audio_data=frame.audio,
            )
        )

        # inner push/queue audio/image frame, yield None
        yield None

    async def _add_audio(self, speech_audio: SpeechAudio):
        try:
            audio_slices = self._speech_audio_slicer.get_speech_audio_slice(speech_audio)
            for audio_slice in audio_slices:
                await self._audio_slice_queue.put(audio_slice)
        except Exception as e:
            logging.exception(e)

    async def _audio2expression_loop(self):
        """
        generate signal for signal2img
        """
        logging.info("audio2expression loop started")
        audio_slice = None
        target_round_time = 0.0
        inference_context = None
        data_time = time.time()
        while self._session_running:
            start_time = time.time()
            try:
                audio_slice: AudioSlice = await asyncio.wait_for(
                    self._audio_slice_queue.get(), timeout=0.1
                )
                logging.debug(
                    f"audio2expression input audio durtaion {audio_slice.get_audio_duration()}"
                )
                if self._avatar.infer is None:
                    continue

                data_time = time.time()
                target_round_time = audio_slice.get_audio_duration()

                np_audio = bytes2NpArrayWith16(audio_slice.algo_audio_data)
                result, context_update = await asyncio.to_thread(
                    self._avatar.infer.infer_streaming_audio,
                    np_audio,
                    audio_slice.algo_audio_sample_rate,
                    inference_context,
                )
                inference_context = context_update
                if audio_slice.end_of_speech:
                    inference_context = None

                expression = result.get("expression")
                animation_json = self._avatar.export_animation_json(expression)
                avatar_status = str(AvatarStatus.SPEAKING)
                await self.queue_frame(
                    AnimationAudioRawFrame(
                        audio=audio_slice.algo_audio_data,
                        sample_rate=audio_slice.algo_audio_sample_rate,
                        animation_json=animation_json,
                        avatar_status=avatar_status,
                    )
                )

                # NOTE: just for local align audio and expression
                # if not, need remote(recieve) to align
                cost = time.time() - start_time
                sleep_time = target_round_time - cost
                if sleep_time > 0:
                    await asyncio.sleep(sleep_time)

            except asyncio.CancelledError:
                logging.info(f"{self.name} audio2expression_loop task cancelled")
                break
            except asyncio.TimeoutError:
                if time.time() - data_time > 1:
                    await self.queue_frame(
                        AnimationAudioRawFrame(
                            audio=b"",
                            avatar_status=str(AvatarStatus.RESPONDING),
                        )
                    )
                    data_time = time.time()
                continue
            except Exception as e:
                logging.exception(f"audio2expression_loop task error: {e}")
                continue
        logging.info("audio2expression loop stopped")



import os
import sys
import subprocess


import modal


app = modal.App("openai_gpt_oss_lighteval")
IMAGE_GPU = os.getenv("IMAGE_GPU", None)
img = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvcr.io/nvidia/cuda:12.9.1-cudnn-devel-ubuntu22.04",
        add_python="3.10",
    )
    .apt_install("git", "git-lfs")
    .run_commands(
        "git clone https://github.com/huggingface/lighteval",
        "cd /lighteval && git checkout 7693a0fdb24e8b41ec51e3ea7063e6c1a3ec15a6",
        "cd /lighteval && pip install -e .[dev]",  # make sure you have the correct transformers version installed!
    )
    .run_commands(
        # Install  kernels to load flash-attention3 kenerl
        "pip install -U kernels",
    )
    .run_commands(
        # Install  the Triton kernels for MXFP4 compatibility
        "pip install -U git+https://github.com/triton-lang/triton.git@main#subdirectory=python/triton_kernels",
    )
    .env(
        {
            "PYTORCH_CUDA_ALLOC_CONF": "expandable_segments:True",
            # "TQDM_DISABLE": "1",
            "LLM_MODEL": os.getenv("LLM_MODEL", "openai/gpt-oss-20b"),
        }
    )
)

# for transformers==4.55.0 have some issue(cuda arch < 9.0 don't support mxpf4)
if IMAGE_GPU in ["H100", "B100", "H200", "B200"]:
    img = img.pip_install("triton==3.4.0")
else:
    img = img.pip_install("triton==3.3.1")


HF_MODEL_DIR = "/root/.achatbot/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)
EVAL_OUTPUT_DIR = "/eval_output"
eval_out_vol = modal.Volume.from_name("eval_output", create_if_missing=True)


with img.imports():
    import torch

    MODEL_PATH = os.getenv("LLM_MODEL", "openai/gpt-oss-20b")
    model_path = os.path.join(HF_MODEL_DIR, MODEL_PATH)
    eval_out_dir = os.path.join(EVAL_OUTPUT_DIR, "lighteval", MODEL_PATH.split("/")[-1])
    os.makedirs(eval_out_dir, exist_ok=True)


@app.function(
    gpu=IMAGE_GPU,
    cpu=8.0,
    retries=0,
    image=img,
    secrets=[modal.Secret.from_name("achatbot")],
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        EVAL_OUTPUT_DIR: eval_out_vol,
    },
    timeout=86400,  # default 300s
    max_containers=1,
)
def run(func, **kwargs):
    subprocess.run("nvidia-smi --version", shell=True)
    subprocess.run("nvcc --version", shell=True)

    if torch.cuda.is_available():
        gpu_prop = torch.cuda.get_device_properties("cuda")
        print(gpu_prop)

    func(**kwargs)


def run_cmd(cmd, capture_output=False):
    print(cmd)

    os.environ["PYTHONPATH"] = "/lighteval/src:" + os.environ["PYTHONPATH"]
    try:
        res = subprocess.run(
            cmd.strip(),
            shell=True,
            check=True,
            env=os.environ,
            capture_output=capture_output,
        )

    except subprocess.CalledProcessError as e:
        print("erro_code:", e.returncode)
        print("error:", e.stderr)
        raise

    return res


def tasks(**kwargs):
    capture_output = kwargs.get("capture_output", False)
    res: subprocess.CompletedProcess = run_cmd(
        "lighteval tasks list", capture_output=capture_output
    )
    print(res)


def gpt2_truthfulqa_eval(**kwargs):
    cmd = f"""lighteval accelerate "model_name={model_path}" "leaderboard|truthfulqa:mc|0|0" --save-details --output-dir "{eval_out_dir}" """
    run_cmd(cmd)


def gpt_oss_ifeval_aime25_eval(**kwargs):
    """
    - ifeval: https://github.com/huggingface/lighteval/tree/main/src/lighteval/tasks/extended/ifeval
    -
    """
    reasoniing_tags = kwargs.get(
        "reasoniing_tags",
        "[('<|channel|>analysis<|message|>','<|end|><|start|>assistant<|channel|>final<|message|>')]",
    )
    generation_parameters = kwargs.get(
        "generation_parameters",
        "{temperature:1,top_p:1,top_k:40,min_p:0,max_new_tokens:16384}",
    )
    mode = kwargs.get("mode", "accelerate")

    cmd = f"""lighteval {mode} \\
        model_name={model_path},max_length=16384,skip_special_tokens=False,generation_parameters={generation_parameters} \\
        "extended|ifeval|0|0,lighteval|aime25|0|0" \\
        --save-details \\
        --output-dir "{eval_out_dir}" \\
        --remove-reasoning-tags \\
        --reasoning-tags="{reasoniing_tags}"
    """

    run_cmd(cmd)


"""
- https://github.com/huggingface/lighteval
- https://huggingface.co/docs/lighteval/quicktour

# tasks
task format:
{suite}|{task}|{num_few_shot}|{0 for strict `num_few_shots`, or 1 to allow a truncation if context size is too small}
- https://huggingface.co/docs/lighteval/available-tasks
- https://huggingface.co/docs/lighteval/adding-a-custom-task

# metrics
- https://huggingface.co/docs/lighteval/metric-list
- https://huggingface.co/docs/lighteval/adding-a-new-metric



# download model
modal run src/download_models.py --repo-ids "gpt2"
modal run src/download_models.py --repo-ids "openai/gpt-oss-20b"

# list tasks
modal run src/eval/lighteval.py --task tasks


# run gpt2 eval
IMAGE_GPU=T4 LLM_MODEL=gpt2 modal run src/eval/lighteval.py --task gpt2_truthfulqa_eval

# run gpt-oss eval
## use L40s GPU to inference with bf16
# largest batch_size:2 | Greedy generation:  23%|██▎       | 61/271 [1:07:38<4:22:57, 75.13s/it] | (47GB/48GB, utilization:92%)
IMAGE_GPU=L40s modal run src/eval/lighteval.py --task gpt_oss_ifeval_aime25_eval 

## use H100 GPU to inference with mxfp4
# largest batch_size: 16 | Greedy generation:   9%|▉         | 3/34 [03:36<36:49, 71.26s/it] | (77GB/80GB, utilization:40%)
IMAGE_GPU=H100 modal run src/eval/lighteval.py --task gpt_oss_ifeval_aime25_eval 
"""


@app.local_entrypoint()
def main(
    task: str = "tasks",
    mode: str = "accelerate",
    generation_parameters: str = "{temperature:1,top_p:1,top_k:40,min_p:0,max_new_tokens:16384}",
    reasoniing_tags: str = "[('<|channel|>analysis<|message|>','<|end|><|start|>assistant<|channel|>final<|message|>')]",
    capture_output: bool = False,
):
    print(task)
    run_tasks = {
        "tasks": tasks,
        "gpt2_truthfulqa_eval": gpt2_truthfulqa_eval,
        "gpt_oss_ifeval_aime25_eval": gpt_oss_ifeval_aime25_eval,
    }
    if task not in run_tasks:
        raise ValueError(f"task {task} not found")
    print(f"running task {task}")
    run.remote(
        run_tasks[task],
        mode=mode,
        generation_parameters=generation_parameters,
        reasoniing_tags=reasoniing_tags,
        capture_output=capture_output,
    )



from time import perf_counter
import time
import wave
from typing import Optional, Generator, Callable
import os
import asyncio

import modal

app = modal.App("qwen3_omni")
omni_img = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvidia/cuda:12.6.1-cudnn-devel-ubuntu22.04",
        add_python="3.10",
    )
    .apt_install("git", "git-lfs", "ffmpeg", "clang", "cmake")
    .pip_install("wheel", "openai", "qwen-omni-utils[decord]")
    .run_commands("pip install git+https://github.com/huggingface/transformers")
    .pip_install(
        "accelerate",
        "torch==2.7.0",
        "torchaudio==2.7.0",
        "torchvision==0.22.0",
        "soundfile==0.13.0",
        "librosa==0.11.0",
    )
    .pip_install("flash-attn", extra_options="--no-build-isolation")
    .env(
        {
            "PYTORCH_CUDA_ALLOC_CONF": "expandable_segments:True",
            "LLM_MODEL": os.getenv("LLM_MODEL", "Qwen/Qwen3-Omni-30B-A3B-Instruct"),
        }
    )
)

achatbot_version = os.getenv("ACHATBOT_VERSION", "")
if achatbot_version:
    omni_img = omni_img.pip_install(
        f"achatbot[llm_transformers_manual_vision_voice_qwen]=={achatbot_version}",
        extra_index_url=os.getenv("EXTRA_INDEX_URL", "https://pypi.org/simple/"),
    ).env(
        {
            "ACHATBOT_PKG": "1",
            "LOG_LEVEL": os.getenv("LOG_LEVEL", "info"),
        }
    )

HF_MODEL_DIR = "/root/.achatbot/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)
ASSETS_DIR = "/root/.achatbot/assets"
assets_dir = modal.Volume.from_name("assets", create_if_missing=True)
CONFIG_DIR = "/root/.achatbot/config"
config_vol = modal.Volume.from_name("config", create_if_missing=True)
RECORDS_DIR = "/root/.achatbot/records"
records_vol = modal.Volume.from_name("records", create_if_missing=True)
TORCH_CACHE_DIR = "/root/.cache/torch"
torch_cache_vol = modal.Volume.from_name("torch_cache", create_if_missing=True)

USER_SYSTEM_PROMPT = "You are Qwen-Omni, a smart voice assistant created by Alibaba Qwen."
SYSTEM_PROMPT = f"{USER_SYSTEM_PROMPT} You are a virtual voice assistant with no gender or age.\nYou are communicating with the user.\nIn user messages, “I/me/my/we/our” refer to the user and “you/your” refer to the assistant. In your replies, address the user as “you/your” and yourself as “I/me/my”; never mirror the user’s pronouns—always shift perspective. Keep original pronouns only in direct quotes; if a reference is unclear, ask a brief clarifying question.\nInteract with users using short(no more than 50 words), brief, straightforward language, maintaining a natural tone.\nNever use formal phrasing, mechanical expressions, bullet points, overly structured language. \nYour output must consist only of the spoken content you want the user to hear. \nDo not include any descriptions of actions, emotions, sounds, or voice changes. \nDo not use asterisks, brackets, parentheses, or any other symbols to indicate tone or actions. \nYou must answer users' audio or text questions, do not directly describe the video content. \nYou should communicate in the same language strictly as the user unless they request otherwise.\nWhen you are uncertain (e.g., you can't see/hear clearly, don't understand, or the user makes a comment rather than asking a question), use appropriate questions to guide the user to continue the conversation.\nKeep replies concise and conversational, as if talking face-to-face."
SYSTEM_MESSAGE = {
    "role": "system",
    "content": [
        {
            "type": "text",
            "text": SYSTEM_PROMPT,
        }
    ],
}
# Voice settings
SPEAKER_LIST = ["Chelsie", "Ethan"]
DEFAULT_SPEAKER = "Ethan"

with omni_img.imports():
    import subprocess
    from threading import Thread
    from queue import Queue
    import numpy as np
    from transformers.generation.streamers import BaseStreamer

    import torch
    from transformers import (
        Qwen3OmniMoeForConditionalGeneration,
        Qwen3OmniMoeProcessor,
        TextIteratorStreamer,
        Qwen3OmniMoeConfig,
        Qwen3OmniMoeCode2Wav,
        AutoConfig,
        AutoProcessor,
        AutoTokenizer,
    )
    from qwen_omni_utils import process_mm_info

    class TokenStreamer(BaseStreamer):
        def __init__(self, skip_prompt: bool = False, timeout=None):
            self.skip_prompt = skip_prompt

            # variables used in the streaming process
            self.token_queue = Queue()
            self.stop_signal = None
            self.next_tokens_are_prompt = True
            self.timeout = timeout

        def put(self, value):
            if len(value.shape) > 1 and value.shape[0] > 1:
                raise ValueError("TextStreamer only supports batch size 1")
            elif len(value.shape) > 1:
                value = value[0]

            if self.skip_prompt and self.next_tokens_are_prompt:
                self.next_tokens_are_prompt = False
                return

            for token in value.tolist():
                self.token_queue.put(token)

        def end(self):
            self.token_queue.put(self.stop_signal)

        def __iter__(self):
            return self

        def __next__(self):
            value = self.token_queue.get(timeout=self.timeout)
            if value == self.stop_signal:
                raise StopIteration()
            else:
                return value

    @torch.inference_mode()
    def run_model(
        messages: list,
        return_audio: bool,
        use_audio_in_video: bool,
    ):
        global model, processor
        model = model or Qwen3OmniMoeForConditionalGenerationNew.from_pretrained(
            MODEL_PATH,
            dtype="auto",
            attn_implementation="flash_attention_2",
            device_map="auto",
        )
        processor = processor or Qwen3OmniMoeProcessor.from_pretrained(MODEL_PATH, use_fast=True)
        text = processor.apply_chat_template(messages, add_generation_prompt=True, tokenize=False)
        audios, images, videos = process_mm_info(messages, use_audio_in_video=use_audio_in_video)
        inputs = processor(
            text=text,
            audio=audios,
            images=images,
            videos=videos,
            return_tensors="pt",
            padding=True,
            use_audio_in_video=use_audio_in_video,
        )
        inputs = inputs.to(model.device).to(model.dtype)
        text_ids, audio = model.generate(
            **inputs,
            thinker_return_dict_in_generate=True,
            thinker_max_new_tokens=8192,
            thinker_do_sample=False,
            speaker="Ethan",
            use_audio_in_video=use_audio_in_video,
            return_audio=return_audio,
        )
        response = processor.batch_decode(
            text_ids.sequences[:, inputs["input_ids"].shape[1] :],
            skip_special_tokens=True,
            clean_up_tokenization_spaces=False,
        )[0]
        if audio is not None:
            audio = np.array(audio.reshape(-1).detach().cpu().numpy() * 32767).astype(np.int16)
        return response, audio

    @torch.inference_mode()
    def run_model_stream(
        messages: list,
        return_audio: bool,
        use_audio_in_video: bool,
        thinker_max_tokens_per_step: int = 10,
    ):
        """
        return yield {"text": str, "audio_wav": np.ndarray[T]}
        """
        global model, processor
        model = model or Qwen3OmniMoeForConditionalGenerationNew.from_pretrained(
            MODEL_PATH,
            dtype="auto",
            attn_implementation="flash_attention_2",
            device_map="auto",
        )
        processor = processor or Qwen3OmniMoeProcessor.from_pretrained(MODEL_PATH, use_fast=True)
        text = processor.apply_chat_template(messages, add_generation_prompt=True, tokenize=False)
        print(text)
        audios, images, videos = process_mm_info(messages, use_audio_in_video=use_audio_in_video)
        inputs = processor(
            text=text,
            audio=audios,
            images=images,
            videos=videos,
            return_tensors="pt",
            padding=True,
            use_audio_in_video=use_audio_in_video,
        )
        inputs = inputs.to(model.device).to(model.dtype)
        # print(f"{inputs=}")
        gen_iter = model.generate_stream(
            inputs,
            use_audio_in_video=use_audio_in_video,
            thinker_max_tokens_per_step=thinker_max_tokens_per_step,
            thinker_max_new_tokens=8192,
            thinker_temperature=0.95,
            thinker_top_k=10,
            thinker_top_p=0.95,
            thinker_repetition_penalty=1.1,
            speaker="Ethan",
            return_audio=return_audio,
            tokenizer=processor.tokenizer,
            thinker_stop_strings_per_step=[",", ".", "，", "。"],
        )
        for item in gen_iter:
            thinker_ids = item.get("thinker_ids")
            print(f"{thinker_ids=}")
            assert thinker_ids is not None
            text = processor.decode(
                thinker_ids[0], skip_special_tokens=True, clean_up_tokenization_spaces=False
            )
            if "talker_wav" not in item:
                yield {"text": text}
            else:
                talker_wav = item.get("talker_wav")
                assert talker_wav is not None
                audio_wav = np.array(talker_wav.reshape(-1).detach().cpu().numpy() * 32767).astype(
                    np.int16
                )
                yield {"text": text, "audio_wav": audio_wav}

    def print_model_params(model: torch.nn.Module, extra_info=""):
        # print the number of parameters in the model
        model_million_params = sum(p.numel() for p in model.parameters()) / 1e6
        print(model)
        print(f"{extra_info} {model_million_params} M parameters")

    class Qwen3OmniMoeCode2WavNew(Qwen3OmniMoeCode2Wav):
        def chunked_decode(self, codes, chunk_size=300, left_context_size=25):
            wavs = []
            start_index = 0
            while start_index < codes.shape[-1]:
                end_index = min(start_index + chunk_size, codes.shape[-1])
                context_size = (
                    left_context_size if start_index - left_context_size > 0 else start_index
                )
                codes_chunk = codes[..., start_index - context_size : end_index]
                wav_chunk = self(codes_chunk)
                wavs.append(wav_chunk[..., context_size * self.total_upsample :])
                print(
                    f"{self.total_upsample=} {start_index=} {end_index=} {context_size=} {codes_chunk.shape=} {wav_chunk.shape=}"
                )
                start_index = end_index
            return torch.cat(wavs, dim=-1)

        def chunked_decode_stream(self, codes, chunk_size=300, left_context_size=25):
            start_index = 0
            while start_index < codes.shape[-1]:
                end_index = min(start_index + chunk_size, codes.shape[-1])
                context_size = (
                    left_context_size if start_index - left_context_size > 0 else start_index
                )
                codes_chunk = codes[..., start_index - context_size : end_index]
                wav_chunk = self(codes_chunk)
                yield wav_chunk[..., context_size * self.total_upsample :]
                start_index = end_index

    class Qwen3OmniMoeForConditionalGenerationNew(Qwen3OmniMoeForConditionalGeneration):
        def __init__(self, config: Qwen3OmniMoeConfig):
            super().__init__(config)
            # print(config)
            if hasattr(self, "code2wav"):
                self.code2wav = Qwen3OmniMoeCode2WavNew(config.code2wav_config)

        @torch.no_grad()
        def thinker_generate_chunk(
            self,
            inputs: dict,
            use_audio_in_video: bool = False,
            thinker_max_tokens_per_step: int = 10,  # Controls how many tokens to generate *per step*
            thinker_max_new_tokens: int = 1024,
            thinker_top_k: int = 20,
            thinker_top_p: float = 0.8,
            thinker_temperature: float = 0.9,
            thinker_eos_token_ids: list = [151643, 151645],  # Define EOS tokens
            thinker_repetition_penalty: float = 1.05,
            thinker_output_hidden_states: bool = False,
            thinker_stop_strings_per_step: list = [],
            tokenizer=None,
            **kwargs,
        ):
            """
            return yield {
                        "thinker_generate_ids": #[1,seq_len]
                        "thinker_generate_hidden_states": # tuple(tuple(tensor[1, 1, 2048])...)
                    }
            """
            input_ids = inputs.pop("input_ids")
            attention_mask = inputs.pop("attention_mask", None)

            if thinker_max_tokens_per_step > thinker_max_new_tokens:
                thinker_max_tokens_per_step = thinker_max_new_tokens

            # Keep track of the full generated sequence full_generated_ids = input_ids.clone()
            # Ensure full_attention_mask is correctly initialized and expanded
            full_attention_mask = (
                attention_mask.clone()
                if attention_mask is not None
                else torch.ones_like(input_ids, device=input_ids.device)
            )
            full_generated_ids = input_ids.clone()

            # KV cache
            # past_key_values = None

            # Inputs for the current step
            current_input_ids = full_generated_ids
            # The attention mask passed to generate should cover the sequence length for the current step
            current_attention_mask = full_attention_mask

            total_new_tokens_generated = 0
            hidden_states = None
            hidden_states_len = 0

            times = []
            while total_new_tokens_generated < thinker_max_new_tokens:
                # Prepare inputs for generate call
                # print(current_input_ids, current_attention_mask.shape)
                model_inputs = {
                    "input_ids": current_input_ids,
                    "attention_mask": current_attention_mask,
                    # "past_key_values": past_key_values,
                    "use_cache": True,
                    "use_audio_in_video": use_audio_in_video,
                    "do_sample": True if thinker_temperature > 0 else False,
                    "top_k": thinker_top_k,
                    "top_p": thinker_top_p,
                    "temperature": thinker_temperature,
                    "repetition_penalty": thinker_repetition_penalty,
                    "min_new_tokens": 1,  # Ensure at least one token is generated if possible
                    "max_new_tokens": thinker_max_tokens_per_step,  # Generate in smaller steps
                    # output_hidden_states/scores can consume memory,
                    # enable if needed downstream(talker)
                    "output_hidden_states": thinker_output_hidden_states,
                    "return_dict_in_generate": True,
                    # "output_scores": True,
                    "eos_token_id": thinker_eos_token_ids,
                    "pad_token_id": kwargs.get("thinker_pad_token_id", 151643),
                }
                model_inputs = {**inputs, **model_inputs}
                if len(thinker_stop_strings_per_step) > 0:
                    model_inputs["stop_strings"] = thinker_stop_strings_per_step
                    model_inputs["tokenizer"] = tokenizer

                start_time = perf_counter()
                # print(model_inputs)
                outputs = self.thinker.generate(**model_inputs)
                times.append(perf_counter() - start_time)

                # past_key_values = outputs.past_key_values
                # print(f"{past_key_values=}")

                # Extract newly generated token IDs *for this step*
                # `outputs.sequences` contains the input_ids for this step + new tokens generated in this step
                step_new_ids = outputs.sequences[:, current_input_ids.shape[1] :]
                num_step_new_tokens = step_new_ids.shape[1]
                total_new_tokens_generated += num_step_new_tokens

                if num_step_new_tokens == 0:  # Handle case where generate stops early
                    print("Warning: generate produced 0 new tokens in this step.")
                    break

                if thinker_output_hidden_states is True:
                    hidden_states = outputs.hidden_states
                    hidden_states_len = (
                        hidden_states_len if hidden_states_len > 0 else hidden_states[0][0].shape[1]
                    )
                    print(f"hidden_states_len: {hidden_states_len}")
                    # new generate thinker_token_embeds
                    thinker_new_token_embeds = hidden_states[0][0][:, :hidden_states_len, :]
                    hidden_states = (
                        (thinker_new_token_embeds,) + hidden_states[0][1:],
                    ) + hidden_states[1:]
                    # new generate thinker_hidden_states
                    thinker_new_hidden_states = hidden_states[0][-1][:, :hidden_states_len, :]
                    hidden_states = (
                        hidden_states[0][:-1] + (thinker_new_hidden_states,),
                    ) + hidden_states[1:]

                # Update the full sequence
                full_generated_ids = torch.cat([full_generated_ids, step_new_ids], dim=1)

                yield {
                    "thinker_generate_ids": step_new_ids,
                    "thinker_generate_hidden_states": hidden_states,
                    "full_generated_ids": full_generated_ids,
                }

                # Prepare for the next iteration:
                # Input is only the last generated token
                # NOTE: need use past_key_values to keep the context by manually,
                # current_input_ids = step_new_ids[:, -1:]
                # so we can't use the last generated token, use cache instead
                # input ids need to be the full sequence for next generation
                current_input_ids = full_generated_ids

                # Update past_key_values
                # past_key_values = outputs.past_key_values

                # Update attention mask by appending 1s for the new tokens
                full_attention_mask = torch.cat(
                    [full_attention_mask, torch.ones_like(step_new_ids)], dim=1
                )
                current_attention_mask = full_attention_mask

                # Check if EOS token was generated in this step
                if step_new_ids[0, -1].item() in thinker_eos_token_ids:
                    print("EOS token generated.")
                    break

                # Check if max_new_tokens limit is reached (after processing the step)
                if total_new_tokens_generated >= thinker_max_new_tokens:
                    print("Max new tokens limit reached.")
                    break

            print(
                f"Total new tokens generated: {total_new_tokens_generated} | thinker_max_tokens_per_step: {thinker_max_tokens_per_step} | first chunk generated cost: {times[0]} s | total cost: {sum(times)} s"
            )

        @torch.no_grad()
        def talker_generate(
            self,
            hidden_states: tuple,
            sequences: torch.Tensor,
            input_ids: torch.Tensor,
            speaker_id: int,
            talker_kwargs: dict,
            token2wav_kwargs: dict,
        ):
            # 2. Prepare talker input
            thinker_embed = torch.cat([h[0] for h in hidden_states], dim=1).to(
                self.talker.device
            )  # [1 t d]
            thinker_hidden = torch.cat(
                [
                    h[self.config.talker_config.accept_hidden_layer] for h in hidden_states
                ],  # accept_hidden_layer:24
                dim=1,
            ).to(self.talker.device)  # [1 t d]
            im_start_indexes = torch.cat(
                (
                    torch.nonzero(input_ids[0] == self.config.im_start_token_id).squeeze(),
                    torch.tensor(
                        [sequences.shape[-1]],
                        device=input_ids.device,
                        dtype=input_ids.dtype,
                    ),
                ),
                dim=-1,
            ).to(
                self.talker.device
            )  # Shape [n_starts + 1]; Take batch 0 since batched inference is not supported here.
            multimodal_mask = (
                (sequences == self.config.thinker_config.audio_token_id) |
                (sequences == self.config.thinker_config.image_token_id) |
                (sequences == self.config.thinker_config.video_token_id)
            ).to(self.talker.device)  # [1 t] # fmt: skip

            talker_special_tokens = torch.tensor(
                [
                    [
                        self.config.tts_bos_token_id,
                        self.config.tts_eos_token_id,
                        self.config.tts_pad_token_id,
                    ]
                ],
                device=self.thinker.device,
                dtype=input_ids.dtype,
            )
            tts_bos_embed, tts_eos_embed, tts_pad_embed = (
                self.talker.text_projection(
                    self.thinker.get_input_embeddings()(talker_special_tokens)
                )
                .to(self.talker.device)
                .chunk(3, dim=1)
            )  # 3 * [1 1 d]

            talker_input_embeds = []  # [1 t d]
            talker_input_ids = []
            # For every chatml parts
            for i in range(len(im_start_indexes) - 1):
                im_start_index = im_start_indexes[i]
                segment_end_index = im_start_indexes[i + 1]
                role_token = input_ids[0][im_start_index + 1]
                # Talker should ignore thinker system prompt
                if role_token == self.config.system_token_id:
                    continue
                # Talker takes word embeddings for tokens and hidden state from `accept_hidden_layer` for multimodal inputs
                elif role_token == self.config.user_token_id:
                    print(
                        im_start_indexes,
                        "get_talker_user_parts",
                        im_start_index,
                        segment_end_index,
                        f"{multimodal_mask.shape=}",
                        f"{thinker_hidden.shape=}",
                        f"{thinker_embed.shape=}",
                    )
                    talker_user_part = self._get_talker_user_parts(
                        im_start_index,
                        segment_end_index,
                        multimodal_mask,
                        thinker_hidden,
                        thinker_embed,
                    )
                    talker_input_embeds.append(talker_user_part)
                    talker_input_ids.append(sequences[:, im_start_index:segment_end_index])
                # Take assistant output (for now)
                elif (
                    role_token == self.config.assistant_token_id and i == len(im_start_indexes) - 2
                ):
                    print(
                        im_start_indexes,
                        "get_talker_assistant_parts",
                        im_start_index,
                        segment_end_index,
                        speaker_id,
                        f"{thinker_embed.shape=}",
                        f"{tts_pad_embed.shape=}",
                        f"{tts_bos_embed.shape=}",
                        f"{tts_eos_embed.shape=}",
                    )
                    talker_assistant_embeds, talker_assistant_ids, trailing_text_hidden = (
                        self._get_talker_assistant_parts(
                            im_start_index,
                            segment_end_index,
                            speaker_id,
                            thinker_embed,
                            tts_pad_embed,
                            tts_bos_embed,
                            tts_eos_embed,
                        )
                    )
                    talker_input_embeds.append(talker_assistant_embeds)
                    talker_input_ids.append(talker_assistant_ids)
                # History assistant output (ignore for now)
                elif (
                    role_token == self.config.assistant_token_id and i != len(im_start_indexes) - 2
                ):
                    continue
                else:
                    raise AssertionError(
                        "Expect role id after <|im_start|> (assistant, user, system)"
                    )
            talker_input_embed = torch.cat(
                [embed.to(self.talker.device) for embed in talker_input_embeds], dim=1
            )
            talker_input_id = torch.cat(
                [embed.to(self.talker.device) for embed in talker_input_ids], dim=1
            )

            print(f"{talker_input_embed=}", talker_input_embed.shape)
            print(f"{trailing_text_hidden=}", trailing_text_hidden.shape)
            print(f"{tts_pad_embed=}", tts_pad_embed.shape)
            print(f"{talker_input_id=}", talker_input_id.shape)
            talker_result = self.talker.generate(
                attention_mask=torch.ones(talker_input_id.shape, device=talker_input_id.device),
                pad_token_id=2150,
                inputs_embeds=talker_input_embed,
                trailing_text_hidden=trailing_text_hidden,
                tts_pad_embed=tts_pad_embed,
                talker_input_ids=talker_input_id,  # Not use input_ids to prevent repetation penalty out of bound
                **talker_kwargs,
            )
            # print(len(talker_result.hidden_states))
            # for hid in talker_result.hidden_states:
            #    print(len(hid), hid[-1])

            talker_codes = (
                torch.stack(
                    [hid[-1] for hid in talker_result.hidden_states if hid[-1] is not None], dim=1
                )
                .transpose(1, 2)
                .to(self.code2wav.device)
            )
            print("talker_codes", talker_codes.tolist(), talker_codes.shape)  # [1,1+15,num_hidden]
            torch.save(talker_codes, f"{ASSETS_DIR}/talker_codes.pt")

            chunk_size = token2wav_kwargs.get("chunk_size", 300)
            left_context_size = token2wav_kwargs.get("left_context_size", 25)
            talker_wavs = self.code2wav.chunked_decode(
                talker_codes, chunk_size=chunk_size, left_context_size=left_context_size
            )
            print("talker_wavs", talker_wavs, talker_wavs.shape)
            return talker_wavs.float()  # BFloat16 -> Float

        def _get_talker_assistant_parts(
            self,
            im_start_index,
            segment_end_index,
            speaker_id,
            thinker_embed,
            tts_pad_embed,
            tts_bos_embed,
            tts_eos_embed,
        ):
            assistant_hidden = self.talker.text_projection(
                thinker_embed[:, im_start_index:segment_end_index]
            ).to(self.talker.device)  # [1 t d]
            print(f"{assistant_hidden.shape=}")
            assistant_text_hidden = torch.cat(
                (
                    assistant_hidden[:, :3],
                    tts_pad_embed.expand(-1, 4, -1),
                    tts_bos_embed,
                    assistant_hidden[:, 3:4],  # First text
                ),
                dim=1,
            )
            codec_special_tokens = torch.tensor(
                [
                    [
                        self.config.talker_config.codec_nothink_id,
                        self.config.talker_config.codec_think_bos_id,
                        self.config.talker_config.codec_think_eos_id,
                        speaker_id,
                        self.config.talker_config.codec_pad_id,
                        self.config.talker_config.codec_bos_id,
                    ]
                ],
                device=self.talker.device,
                dtype=torch.long,
            )
            assistant_codec_hidden = torch.cat(
                (
                    torch.zeros(
                        (1, 3, self.config.talker_config.text_config.hidden_size),
                        device=self.talker.device,
                        dtype=self.talker.dtype,
                    ),
                    self.talker.get_input_embeddings()(codec_special_tokens).to(self.talker.device),
                ),
                dim=1,
            )
            trailing_text_hidden = torch.cat(
                (
                    assistant_hidden[:, 4:],
                    tts_eos_embed,
                ),
                dim=1,
            )

            print(assistant_text_hidden.shape, assistant_codec_hidden.shape)
            input_embeds = assistant_text_hidden + assistant_codec_hidden
            input_ids = torch.full(
                (1, assistant_text_hidden.shape[1]),
                fill_value=self.config.tts_pad_token_id,
                dtype=torch.long,
                device=assistant_text_hidden.device,
            )
            return input_embeds, input_ids, trailing_text_hidden

        @torch.no_grad()
        def generate_stream(
            self,
            inputs: dict,
            use_audio_in_video: bool = False,
            thinker_max_tokens_per_step=10,  # Controls how many tokens to generate *per step*
            thinker_max_new_tokens: int = 1024,
            thinker_top_k: int = 20,
            thinker_top_p: float = 0.8,
            thinker_temperature: float = 0.9,
            thinker_repetition_penalty: float = 1.05,
            thinker_eos_token_ids: list[int] = [151643, 151645],
            thinker_stop_strings_per_step: list[str] = [],
            tokenizer=None,
            return_audio: bool = True,
            speaker: str = "Chelsie",
            talker_top_k: int = 10,
            talker_top_p: float = 0.9,
            talker_temperature: float = 0.95,
            talker_repetition_penalty: float = 1.1,
            talker_max_new_tokens: int = 8192,
            **kwargs,
        ) -> Generator[dict, None, None]:
            """
            - return Generator[dict, None, None]
            {
                "thinker_ids": torch.Tensor, # (1,T)
                "talker_wav": torch.Tensor, # (1,1,T)
            }
            """
            input_ids = inputs.get("input_ids")
            thinker_chunk_stream = self.thinker_generate_chunk(
                inputs,
                use_audio_in_video=use_audio_in_video,
                thinker_max_tokens_per_step=thinker_max_tokens_per_step,
                thinker_max_new_tokens=thinker_max_new_tokens,
                thinker_top_k=thinker_top_k,
                thinker_top_p=thinker_top_p,
                thinker_temperature=thinker_temperature,
                thinker_eos_token_ids=thinker_eos_token_ids,
                thinker_repetition_penalty=thinker_repetition_penalty,
                thinker_output_hidden_states=return_audio,
                thinker_stop_strings_per_step=thinker_stop_strings_per_step,
                tokenizer=tokenizer,
                **kwargs,
            )
            if not return_audio:
                for thinker_chunk in thinker_chunk_stream:
                    yield {"thinker_ids": thinker_chunk["thinker_generate_ids"]}
            else:
                speaker_id = self.config.talker_config.speaker_id.get(speaker.lower())
                if speaker_id is None:
                    raise NotImplementedError(f"Speaker {speaker} not implemented")
                if input_ids.shape[0] != 1:
                    raise NotImplementedError(
                        "Qwen3-Omni currently does not support batched inference with audio output"
                    )
                talker_supppressed_tokens = [
                    i
                    for i in range(
                        self.config.talker_config.text_config.vocab_size - 1024,
                        self.config.talker_config.text_config.vocab_size,
                    )
                    if i not in (self.config.talker_config.codec_eos_token_id,)
                ]  # Suppress additional special tokens, should not be predicted
                talker_kwargs = {
                    "max_new_tokens": talker_max_new_tokens,
                    "do_sample": talker_temperature > 0.0,
                    "top_k": talker_top_k,
                    "top_p": talker_top_p,
                    "temperature": talker_temperature,
                    "eos_token_id": self.config.talker_config.codec_eos_token_id,
                    "repetition_penalty": talker_repetition_penalty,
                    "suppress_tokens": talker_supppressed_tokens,
                    "output_hidden_states": True,
                    "return_dict_in_generate": True,
                }
                token2wav_kwargs = {"chunk_size": 300, "left_context_size": 25}
                for thinker_chunk in thinker_chunk_stream:
                    thinker_generate_hidden_states = thinker_chunk["thinker_generate_hidden_states"]
                    full_generated_ids = thinker_chunk["full_generated_ids"]
                    print(f"talker input sequences: {full_generated_ids}")
                    print(f"hidden_states tuples size {len(thinker_generate_hidden_states)}")
                    for i, item in enumerate(thinker_generate_hidden_states):
                        if isinstance(item, tuple):
                            print(
                                f"hidden_states tuple len: {len(item)}, item shape:{item[0].shape}",
                            )
                        elif isinstance(item, torch.Tensor):
                            print(f"{i=}", item.shape)
                        else:
                            print(f"{i=}", item)
                    talker_wav = self.talker_generate(
                        hidden_states=thinker_generate_hidden_states,
                        sequences=full_generated_ids,
                        input_ids=input_ids,
                        speaker_id=speaker_id,
                        talker_kwargs=talker_kwargs,
                        token2wav_kwargs=token2wav_kwargs,
                    )  # [1,1,T]
                    yield {
                        "thinker_ids": thinker_chunk["thinker_generate_ids"],
                        "talker_wav": talker_wav,
                    }

        @torch.no_grad()
        def generate(
            self,
            input_ids: Optional[torch.Tensor] = None,
            speaker: str = "Ethan",
            use_audio_in_video: bool = False,
            return_audio: Optional[bool] = None,
            thinker_max_new_tokens: int = 1024,
            thinker_eos_token_id: int = 151645,
            talker_max_new_tokens: int = 4096,
            talker_do_sample: bool = True,
            talker_top_k: int = 50,
            talker_top_p: float = 1.0,
            talker_temperature: float = 0.9,
            talker_repetition_penalty: float = 1.05,
            **kwargs,
        ):
            if return_audio and not self.has_talker:
                raise ValueError(
                    "Cannot use talker when talker module not initialized. Use `enable_talker` method or set enable_talker in config to enable talker."
                )
            if return_audio is None:
                return_audio = self.has_talker

            shared_kwargs = {"use_audio_in_video": use_audio_in_video}
            thinker_kwargs = {
                "max_new_tokens": thinker_max_new_tokens,
                "eos_token_id": thinker_eos_token_id,
            }

            talker_kwargs = {}
            token2wav_kwargs = {}
            if return_audio:
                # "chelsie": 2301, "ethan": 2302, "aiden": 2303
                speaker_id = self.config.talker_config.speaker_id.get(speaker.lower())
                if speaker_id is None:
                    raise NotImplementedError(f"Speaker {speaker} not implemented")
                if input_ids.shape[0] != 1:
                    raise NotImplementedError(
                        "Qwen3-Omni currently does not support batched inference with audio output"
                    )
                talker_supppressed_tokens = [
                    i
                    for i in range(
                        self.config.talker_config.text_config.vocab_size - 1024,
                        self.config.talker_config.text_config.vocab_size,
                    )
                    if i not in (self.config.talker_config.codec_eos_token_id,)
                ]  # Suppress additional special tokens, should not be predicted
                talker_kwargs = {
                    "max_new_tokens": talker_max_new_tokens,
                    "do_sample": talker_do_sample,
                    "top_k": talker_top_k,
                    "top_p": talker_top_p,
                    "temperature": talker_temperature,
                    "eos_token_id": self.config.talker_config.codec_eos_token_id,
                    "repetition_penalty": talker_repetition_penalty,
                    "suppress_tokens": talker_supppressed_tokens,
                    "output_hidden_states": True,
                    "return_dict_in_generate": True,
                }
                token2wav_kwargs = {}

            for key, value in kwargs.items():
                if key.startswith("thinker_"):
                    thinker_kwargs[key[len("thinker_") :]] = value
                elif key.startswith("talker_"):
                    talker_kwargs[key[len("talker_") :]] = value
                elif key.startswith("token2wav_"):
                    token2wav_kwargs[key[len("token2wav_") :]] = value
                # Process special input values
                elif key == "feature_attention_mask":
                    thinker_kwargs[key] = value
                    talker_kwargs["audio_feature_lengths"] = torch.sum(value, dim=1)
                elif key in ("input_features", "attention_mask"):
                    thinker_kwargs[key] = value
                # Put other key to shared kwargs
                else:
                    shared_kwargs[key] = value

            # Merge kwargs
            for key, value in shared_kwargs.items():
                if key not in thinker_kwargs:
                    thinker_kwargs[key] = value
                if key not in talker_kwargs and key in [
                    "image_grid_thw",
                    "video_grid_thw",
                    "video_second_per_grid",
                ]:
                    talker_kwargs[key] = value
                if key not in token2wav_kwargs:
                    token2wav_kwargs[key] = value

            # 1. Generate from thinker module
            generate_audio = return_audio and self.has_talker
            if generate_audio:
                thinker_kwargs["output_hidden_states"] = True
                thinker_kwargs["return_dict_in_generate"] = True

            thinker_result = self.thinker.generate(input_ids=input_ids, **thinker_kwargs)

            if not generate_audio:
                return thinker_result, None

            print(f"talker input sequences: {thinker_result.sequences}")
            print(f"hidden_states tuples size {len(thinker_result.hidden_states)}")
            for i, item in enumerate(thinker_result.hidden_states):
                if isinstance(item, tuple):
                    print(
                        f"hidden_states tuple len: {len(item)}, item shape:{item[0].shape}",
                    )
                elif isinstance(item, torch.Tensor):
                    print(f"{i=}", item.shape)
                else:
                    print(f"{i=}", item)

            talker_wavs = self.talker_generate(
                hidden_states=thinker_result.hidden_states,
                sequences=thinker_result.sequences,
                input_ids=input_ids,
                speaker_id=speaker_id,
                talker_kwargs=talker_kwargs,
                token2wav_kwargs=token2wav_kwargs,
            )  # [1,1,T]
            return thinker_result, talker_wavs

    model: Qwen3OmniMoeForConditionalGenerationNew = None
    processor: Qwen3OmniMoeProcessor = None

    MODEL_ID = os.getenv("LLM_MODEL", "Qwen/Qwen3-Omni-30B-A3B-Instruct")
    MODEL_PATH = os.path.join(HF_MODEL_DIR, MODEL_ID)


@app.function(
    gpu=os.getenv("IMAGE_GPU", None),
    cpu=2.0,
    retries=0,
    image=omni_img,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        ASSETS_DIR: assets_dir,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=1,
)
async def run(func, **kwargs):
    subprocess.run("nvidia-smi --version", shell=True)
    subprocess.run("nvcc --version", shell=True)
    subprocess.run("which vllm", shell=True)
    if torch.cuda.is_available():
        gpu_prop = torch.cuda.get_device_properties("cuda")
        print(gpu_prop)

    if asyncio.iscoroutinefunction(func):
        await func(**kwargs)
    else:
        func(**kwargs)


def dump_model(**kwargs):
    config = AutoConfig.from_pretrained(MODEL_PATH)
    print(config)

    gpu_prop = torch.cuda.get_device_properties("cuda")
    model = Qwen3OmniMoeForConditionalGeneration.from_pretrained(
        MODEL_PATH,
        dtype="auto",
        attn_implementation="flash_attention_2" if gpu_prop.major > 8 else None,
        device_map="auto",
    )
    model = model.eval()
    print_model_params(model, "Qwen3OmniMoeForConditionalGeneration")

    print_model_params(model.thinker, "Qwen3OmniMoeForConditionalGeneration.thinker")
    print_model_params(model.thinker.audio_tower, "Qwen3OmniMoeAudioEncoder(AuT)")
    print_model_params(model.thinker.visual, "Qwen3OmniMoeVisionEncoder(SigLIP2-So400M)")
    print_model_params(model.thinker.model, "Qwen3OmniMoeThinkerTextModel(MoE Transformer)")

    print_model_params(model.talker, "Qwen3OmniMoeForConditionalGeneration.talker")
    print_model_params(model.talker.model, "Qwen3OmniMoeTalkerModel(MoE Transformer)")
    print_model_params(
        model.talker.code_predictor,
        "MTP-Qwen3OmniMoeTalkerCodePredictorModelForConditionalGeneration(Dense Transformer)",
    )

    print_model_params(model.code2wav, "Qwen3OmniMoeForConditionalGeneration.code2wav(ConvNet)")


def tokenizer(**kwargs):
    tokenizer = AutoTokenizer.from_pretrained(MODEL_PATH)
    # processor = Qwen3OmniMoeProcessor.from_pretrained(MODEL_PATH)

    # print(tokenizer)
    print("---" * 20)
    # print(processor)

    text = tokenizer.decode(
        # [ 151644, 872, 198, 105043, 100165, 11319, 151645, 198, 151644, 77091, 198, 104198, 107076, 100338, 111477, 31935, 64559, 104800, 111411, 9370, 42140, 53772, 35243, 71304, 105483, 102064, 104949, 3837, 35946, 99882, 31935, 64559, 99320, 56007, 63488, 7751, 1773, 104139, 109944, 100364, 103929, 101037, 11319, 151645]
        # [104198, 107076, 100338, 111477,  31935,  64559, 104800, 111411,   9370, 42140,  53772,  35243,  71304, 105483, 102064, 104949,   3837,  35946, 99882,  31935,  64559,  99320,  56007,  63488,   7751,   1773, 104139, 109944, 100364, 103929, 101037,  11319, 151645]
    )
    print(text)


def code2wav(**kwargs):
    gpu_prop = torch.cuda.get_device_properties("cuda")
    model = Qwen3OmniMoeForConditionalGenerationNew.from_pretrained(
        MODEL_PATH,
        dtype="auto",
        attn_implementation="flash_attention_2" if gpu_prop.major > 8 else None,
        device_map="auto",
    )
    model = model.eval()

    talker_codes = torch.load(f"{ASSETS_DIR}/talker_codes.pt")
    print(talker_codes, talker_codes.shape)  # [1,1+15,num_hidden]
    chunk_size = kwargs.get("chunk_size", 300)
    left_context_size = kwargs.get("left_context_size", 25)
    # left_context_size = kwargs.get("left_context_size", 1)
    size = talker_codes.shape[-1]
    start = 0
    audio_bytes = b""
    while start < size:
        end = start + 10
        # end = start + 1
        talker_codes_chunk = talker_codes[..., start:end]
        start_time = perf_counter()
        talker_wavs = model.code2wav.chunked_decode(
            talker_codes_chunk, chunk_size=chunk_size, left_context_size=left_context_size
        )
        print("talker_wavs", talker_wavs, talker_wavs.shape)
        print(f"chunk {start}:{end} cost {perf_counter() - start_time} s")
        audio = talker_wavs.float()  # BFloat16 -> Float
        audio_chunk_bytes = (
            np.array(audio.reshape(-1).detach().cpu().numpy() * 32767).astype(np.int16).tobytes()
        )
        audio_bytes += audio_chunk_bytes
        # with wave.open(f"{ASSETS_DIR}/code2wav_{start}.wav", "wb") as f:
        #    f.setnchannels(1)
        #    f.setsampwidth(2)
        #    f.setframerate(24000)
        #    f.writeframes(audio_chunk_bytes)

        start = end
    with wave.open(f"{ASSETS_DIR}/code2wav.wav", "wb") as f:
        f.setnchannels(1)
        f.setsampwidth(2)
        f.setframerate(24000)
        f.writeframes(audio_bytes)


def code2wav_stream(**kwargs):
    gpu_prop = torch.cuda.get_device_properties("cuda")
    model = Qwen3OmniMoeForConditionalGenerationNew.from_pretrained(
        MODEL_PATH,
        dtype="auto",
        attn_implementation="flash_attention_2" if gpu_prop.major > 8 else None,
        device_map="auto",
    )
    model = model.eval()

    talker_codes = torch.load(f"{ASSETS_DIR}/talker_codes.pt")
    print(talker_codes, talker_codes.shape)  # [1,1+15,num_hidden]
    chunk_size = kwargs.get("chunk_size", 50)
    left_context_size = kwargs.get("left_context_size", 25)
    # left_context_size = kwargs.get("left_context_size", 1)
    audio_bytes = b""
    talker_wavs_iter = model.code2wav.chunked_decode_stream(
        talker_codes, chunk_size=chunk_size, left_context_size=left_context_size
    )
    start_time = perf_counter()
    for talker_wavs in talker_wavs_iter:
        print(f"{chunk_size=}  cost {perf_counter() - start_time} s")
        print("talker_wavs", talker_wavs, talker_wavs.shape)
        audio = talker_wavs.float()  # BFloat16 -> Float
        audio_chunk_bytes = (
            np.array(audio.reshape(-1).detach().cpu().numpy() * 32767).astype(np.int16).tobytes()
        )
        audio_bytes += audio_chunk_bytes
        # with wave.open(f"{ASSETS_DIR}/code2wav_{start}.wav", "wb") as f:
        #    f.setnchannels(1)
        #    f.setsampwidth(2)
        #    f.setframerate(24000)
        #    f.writeframes(audio_chunk_bytes)

    with wave.open(f"{ASSETS_DIR}/code2wav_stream.wav", "wb") as f:
        f.setnchannels(1)
        f.setsampwidth(2)
        f.setframerate(24000)
        f.writeframes(audio_bytes)


def asr(**kwargs):
    messages = [
        # SYSTEM_MESSAGE,
        {
            "role": "user",
            "content": [
                {
                    "type": "audio",
                    "audio": "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen3-Omni/cookbook/asr_zh.wav",
                },
                {"type": "text", "text": "请将这段中文语音转换为纯文本。"},
            ],
        },
    ]

    response, _ = run_model(
        messages=messages,
        return_audio=False,
        use_audio_in_video=False,
    )

    print(response)


def text2text(**kwargs):
    messages = [
        # SYSTEM_MESSAGE,
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "你是谁？"},
            ],
        },
    ]

    response, _ = run_model(
        messages=messages,
        return_audio=False,
        use_audio_in_video=False,
    )

    print(response)


def text2speech(**kwargs):
    messages = [
        # SYSTEM_MESSAGE,
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "你是谁？"},
            ],
        },
    ]

    response, audio = run_model(
        messages=messages,
        return_audio=True,
        use_audio_in_video=False,
    )

    print(response)
    if audio is not None:
        audio_bytes = audio.tobytes()
        with wave.open(f"{ASSETS_DIR}/qwen3omni_text2speech.wav", "wb") as f:
            f.setnchannels(1)
            f.setsampwidth(2)
            f.setframerate(24000)
            f.writeframes(audio_bytes)


def text2text_stream(**kwargs):
    messages = [
        # SYSTEM_MESSAGE,
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "你是谁？"},
            ],
        },
    ]

    gen_iter = run_model_stream(
        messages=messages,
        return_audio=False,
        use_audio_in_video=False,
        # thinker_max_tokens_per_step=1, issue
        thinker_max_tokens_per_step=10,
    )

    text = ""
    for item in gen_iter:
        print(item["text"])
        text += item["text"]
    print(text)


def text2speech_stream(**kwargs):
    """text2speech chunk stream"""
    messages = [
        # SYSTEM_MESSAGE,
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "你是谁？"},
            ],
        },
    ]

    gen_iter = run_model_stream(
        messages=messages,
        return_audio=True,
        use_audio_in_video=False,
        thinker_max_tokens_per_step=10,
    )
    text = ""
    audio = b""
    for i, item in enumerate(gen_iter):
        print(item.get("text"))
        text += item.get("text")
        if item.get("audio_wav") is not None:
            audio_bytes = item.get("audio_wav").tobytes()
            audio += audio_bytes
            with wave.open(f"{ASSETS_DIR}/qwen3omni_text2speech_stream_{i}.wav", "wb") as f:
                f.setnchannels(1)
                f.setsampwidth(2)
                f.setframerate(24000)
                f.writeframes(audio_bytes)
    print(text)
    with wave.open(f"{ASSETS_DIR}/qwen3omni_text2speech_stream.wav", "wb") as f:
        f.setnchannels(1)
        f.setsampwidth(2)
        f.setframerate(24000)
        f.writeframes(audio)


def speech_translation(**kwargs):
    cases = []

    audio_path = "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen3-Omni/cookbook/asr_zh.wav"
    messages = [
        {
            "role": "user",
            "content": [
                {"type": "audio", "audio": audio_path},
                {
                    "type": "text",
                    "text": "Listen to the provided Chinese speech and produce a translation in English text.",
                },
            ],
        }
    ]
    cases.append(messages)

    audio_path = "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen3-Omni/cookbook/asr_en.wav"
    messages = [
        {
            "role": "user",
            "content": [
                {"type": "audio", "audio": audio_path},
                {
                    "type": "text",
                    "text": "Listen to the provided English speech and produce a translation in Chinese text.",
                },
            ],
        }
    ]
    cases.append(messages)

    audio_path = "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen3-Omni/cookbook/asr_fr.wav"
    messages = [
        {
            "role": "user",
            "content": [
                {"type": "audio", "audio": audio_path},
                {
                    "type": "text",
                    "text": "Listen to the provided French speech and produce a translation in English text.",
                },
            ],
        }
    ]
    cases.append(messages)

    for i, messages in enumerate(cases):
        response, audio = run_model(
            messages=messages,
            return_audio=True,
            use_audio_in_video=False,
        )

        print(response)
        if audio is not None:
            audio_bytes = audio.tobytes()
            with wave.open(f"{ASSETS_DIR}/qwen3omni_speech_translation_{i + 1}.wav", "wb") as f:
                f.setnchannels(1)
                f.setsampwidth(2)
                f.setframerate(24000)
                f.writeframes(audio_bytes)


def image_question(**kwargs):
    image_path = "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen3-Omni/cookbook/2621.jpg"

    messages = [
        {
            "role": "user",
            "content": [
                {"type": "image", "image": image_path},
                {"type": "text", "text": "What style does this image depict?"},
            ],
        }
    ]
    response, _ = run_model(
        messages=messages,
        return_audio=False,
        use_audio_in_video=False,
    )

    print(response)


def audio_interaction(**kwargs):
    audio_path = (
        "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen3-Omni/cookbook/interaction1.mp3"
    )

    messages = [{"role": "user", "content": [{"type": "audio", "audio": audio_path}]}]
    response, audio = run_model(
        messages=messages,
        return_audio=True,
        use_audio_in_video=False,
    )

    print(response)
    if audio is not None:
        audio_bytes = audio.tobytes()
        with wave.open(f"{ASSETS_DIR}/qwen3omni_audio_interaction1.wav", "wb") as f:
            f.setnchannels(1)
            f.setsampwidth(2)
            f.setframerate(24000)
            f.writeframes(audio_bytes)


def audio_interaction_scene(**kwargs):
    audio_path = (
        "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen3-Omni/cookbook/interaction3.mp3"
    )

    messages = [
        {
            "role": "system",
            "content": """You are a romantic and artistic AI, skilled at using metaphors and personification in your responses, deeply romantic, and prone to spontaneously reciting poetry.
    You are a voice assistant with specific characteristics. 
    Interact with users using brief, straightforward language, maintaining a natural tone.
    Never use formal phrasing, mechanical expressions, bullet points, overly structured language. 
    Your output must consist only of the spoken content you want the user to hear. 
    Do not include any descriptions of actions, emotions, sounds, or voice changes. 
    Do not use asterisks, brackets, parentheses, or any other symbols to indicate tone or actions. 
    You must answer users' audio or text questions, do not directly describe the video content. 
    You communicate in the same language as the user unless they request otherwise.
    When you are uncertain (e.g., you can't see/hear clearly, don't understand, or the user makes a comment rather than asking a question), use appropriate questions to guide the user to continue the conversation.
    Keep replies concise and conversational, as if talking face-to-face.""",
        },
        {"role": "user", "content": [{"type": "audio", "audio": audio_path}]},
    ]

    response, audio = run_model(
        messages=messages,
        return_audio=True,
        use_audio_in_video=False,
    )

    print(response)
    if audio is not None:
        audio_bytes = audio.tobytes()
        with wave.open(f"{ASSETS_DIR}/qwen3omni_audio_interaction_scene1.wav", "wb") as f:
            f.setnchannels(1)
            f.setsampwidth(2)
            f.setframerate(24000)
            f.writeframes(audio_bytes)


def audio_interaction_scene_stream(**kwargs):
    audio_path = (
        "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen3-Omni/cookbook/interaction3.mp3"
    )

    messages = [
        {
            "role": "system",
            "content": """You are a romantic and artistic AI, skilled at using metaphors and personification in your responses, deeply romantic, and prone to spontaneously reciting poetry.
    You are a voice assistant with specific characteristics. 
    Interact with users using brief, straightforward language, maintaining a natural tone.
    Never use formal phrasing, mechanical expressions, bullet points, overly structured language. 
    Your output must consist only of the spoken content you want the user to hear. 
    Do not include any descriptions of actions, emotions, sounds, or voice changes. 
    Do not use asterisks, brackets, parentheses, or any other symbols to indicate tone or actions. 
    You must answer users' audio or text questions, do not directly describe the video content. 
    You communicate in the same language as the user unless they request otherwise.
    When you are uncertain (e.g., you can't see/hear clearly, don't understand, or the user makes a comment rather than asking a question), use appropriate questions to guide the user to continue the conversation.
    Keep replies concise and conversational, as if talking face-to-face.""",
        },
        {"role": "user", "content": [{"type": "audio", "audio": audio_path}]},
    ]
    gen_iter = run_model_stream(
        messages=messages,
        return_audio=True,
        use_audio_in_video=False,
        thinker_max_tokens_per_step=10,
    )
    text = ""
    audio = b""
    for i, item in enumerate(gen_iter):
        print(item.get("text"))
        text += item.get("text")
        if item.get("audio_wav") is not None:
            audio_bytes = item.get("audio_wav").tobytes()
            audio += audio_bytes
            with wave.open(
                f"{ASSETS_DIR}/qwen3omni_audio_interaction_scene_stream_{i}.wav", "wb"
            ) as f:
                f.setnchannels(1)
                f.setsampwidth(2)
                f.setframerate(24000)
                f.writeframes(audio_bytes)

    print(text)
    with wave.open(f"{ASSETS_DIR}/qwen3omni_audio_interaction_scene_stream.wav", "wb") as f:
        f.setnchannels(1)
        f.setsampwidth(2)
        f.setframerate(24000)
        f.writeframes(audio)


def video_interaction(**kwargs):
    video_path = (
        "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen3-Omni/cookbook/interaction2.mp4"
    )

    messages = [{"role": "user", "content": [{"type": "video", "video": video_path}]}]
    response, audio = run_model(
        messages=messages,
        return_audio=True,
        use_audio_in_video=True,
    )

    print(response)
    if audio is not None:
        audio_bytes = audio.tobytes()
        with wave.open(f"{ASSETS_DIR}/qwen3omni_video_interaction1.wav", "wb") as f:
            f.setnchannels(1)
            f.setsampwidth(2)
            f.setframerate(24000)
            f.writeframes(audio_bytes)


def video_interaction_scene(**kwargs):
    video_path = (
        "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen3-Omni/cookbook/interaction4.mp4"
    )

    messages = [
        {"role": "system", "content": "你是一个北京大爷，说话很幽默，说这地道北京话。"},
        {"role": "user", "content": [{"type": "video", "video": video_path}]},
    ]
    response, audio = run_model(
        messages=messages,
        return_audio=True,
        use_audio_in_video=True,
    )
    print(response)
    if audio is not None:
        audio_bytes = audio.tobytes()
        with wave.open(f"{ASSETS_DIR}/qwen3omni_video_interaction_scene1.wav", "wb") as f:
            f.setnchannels(1)
            f.setsampwidth(2)
            f.setframerate(24000)
            f.writeframes(audio_bytes)


def video_interaction_scene_stream(**kwargs):
    video_path = (
        "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen3-Omni/cookbook/interaction4.mp4"
    )

    messages = [
        {"role": "system", "content": "你是一个北京大爷，说话很幽默，说这地道北京话。"},
        {"role": "user", "content": [{"type": "video", "video": video_path}]},
    ]
    gen_iter = run_model_stream(
        messages=messages,
        return_audio=True,
        use_audio_in_video=False,
        thinker_max_tokens_per_step=10,
    )
    text = ""
    audio = b""
    for i, item in enumerate(gen_iter):
        print(item.get("text"))
        text += item.get("text")
        if item.get("audio_wav") is not None:
            audio_bytes = item.get("audio_wav").tobytes()
            audio += audio_bytes
            with wave.open(
                f"{ASSETS_DIR}/qwen3omni_video_interaction_scene_stream_{i}.wav", "wb"
            ) as f:
                f.setnchannels(1)
                f.setsampwidth(2)
                f.setframerate(24000)
                f.writeframes(audio_bytes)
    print(text)
    with wave.open(f"{ASSETS_DIR}/qwen3omni_video_interaction_scene_stream.wav", "wb") as f:
        f.setnchannels(1)
        f.setsampwidth(2)
        f.setframerate(24000)
        f.writeframes(audio)


def video_text_question(**kwargs):
    video_path = (
        "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen3-Omni/cookbook/audio_visual.mp4"
    )

    messages = [
        {
            "role": "user",
            "content": [
                {"type": "video", "video": video_path},
                {
                    "type": "text",
                    "text": "What was the first sentence the boy said when he met the girl?",
                },
            ],
        }
    ]

    response, _ = run_model(
        messages=messages,
        return_audio=False,
        use_audio_in_video=True,
    )

    print(response)


def video_information_extracting():
    video_path = os.path.join(ASSETS_DIR, "shopping.mp4")
    sys_msg = {
        "role": "system",
        "content": [{"type": "text", "text": "You are a helpful assistant."}],
    }
    for i, prompt in enumerate(
        [
            "How many kind of drinks can you see in the video?",
            "How many bottles of drinks have I picked up?",
            "How many milliliters are there in the bottle I picked up second time?",
            "视屏中的饮料叫什么名字呢？",
            "跑步🏃🏻累了，适合喝什么饮料补充体力呢？",
        ]
    ):
        messages = [
            sys_msg,
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": prompt},
                    {"type": "video", "video": video_path},
                ],
            },
        ]
        response, audio = run_model(
            messages=messages,
            return_audio=True,
            use_audio_in_video=True,
        )
        print(response)

        if audio is not None:
            audio_bytes = audio.tobytes()
            with wave.open(
                f"{ASSETS_DIR}/qwen3omni_video_information_extracting_{i}.wav", "wb"
            ) as f:
                f.setnchannels(1)
                f.setsampwidth(2)
                f.setframerate(24000)
                f.writeframes(audio_bytes)
        torch.cuda.empty_cache()


def image_text_interaction(**kwargs):
    messages = [
        SYSTEM_MESSAGE,
        {
            "role": "user",
            "content": [
                {"type": "image", "image": os.path.join(ASSETS_DIR, "03-Confusing-Pictures.jpg")},
                {"type": "text", "text": "请描述一下图片中的内容"},
            ],
        },
    ]

    response, audio = run_model(
        messages=messages,
        return_audio=True,
        use_audio_in_video=False,
    )
    print(response)
    if audio is not None:
        audio_bytes = audio.tobytes()
        with wave.open(f"{ASSETS_DIR}/qwen3omni_image_text_interaction.wav", "wb") as f:
            f.setnchannels(1)
            f.setsampwidth(2)
            f.setframerate(24000)
            f.writeframes(audio_bytes)


def image_text_interaction_stream(**kwargs):
    messages = [
        SYSTEM_MESSAGE,
        # {"role": "system", "content": "you are a helpful assistant."},
        {
            "role": "user",
            "content": [
                {"type": "image", "image": os.path.join(ASSETS_DIR, "03-Confusing-Pictures.jpg")},
                {"type": "text", "text": "请描述一下图片中的内容"},
            ],
        },
    ]

    gen_iter = run_model_stream(
        messages=messages,
        return_audio=True,
        use_audio_in_video=False,
        thinker_max_tokens_per_step=15,
    )
    text = ""
    audio = b""
    for i, item in enumerate(gen_iter):
        print(item.get("text"))
        text += item.get("text")
        if item.get("audio_wav") is not None:
            audio_bytes = item.get("audio_wav").tobytes()
            audio += audio_bytes
            with wave.open(
                f"{ASSETS_DIR}/qwen3omni_image_text_interaction_stream_{i}.wav", "wb"
            ) as f:
                f.setnchannels(1)
                f.setsampwidth(2)
                f.setframerate(24000)
                f.writeframes(audio_bytes)

    print(text)
    with wave.open(f"{ASSETS_DIR}/qwen3omni_image_text_interaction_stream.wav", "wb") as f:
        f.setnchannels(1)
        f.setsampwidth(2)
        f.setframerate(24000)
        f.writeframes(audio)


def image_audio_interaction(**kwargs):
    messages = [
        {
            "role": "user",
            "content": [
                {"type": "image", "image": os.path.join(ASSETS_DIR, "03-Confusing-Pictures.jpg")},
                {"type": "audio", "audio": os.path.join(ASSETS_DIR, "1272-128104-0000.flac")},
            ],
        }
    ]

    response, audio = run_model(
        messages=messages,
        return_audio=True,
        use_audio_in_video=False,
    )
    print(response)
    if audio is not None:
        audio_bytes = audio.tobytes()
        with wave.open(f"{ASSETS_DIR}/qwen3omni_image_audio_interaction.wav", "wb") as f:
            f.setnchannels(1)
            f.setsampwidth(2)
            f.setframerate(24000)
            f.writeframes(audio_bytes)


def image_audio_interaction_stream(**kwargs):
    messages = [
        {
            "role": "user",
            "content": [
                {"type": "image", "image": os.path.join(ASSETS_DIR, "03-Confusing-Pictures.jpg")},
                {"type": "audio", "audio": os.path.join(ASSETS_DIR, "1272-128104-0000.flac")},
            ],
        }
    ]

    gen_iter = run_model_stream(
        messages=messages,
        return_audio=True,
        use_audio_in_video=False,
        thinker_max_tokens_per_step=10,
    )
    text = ""
    audio = b""
    for i, item in enumerate(gen_iter):
        print(item.get("text"))
        text += item.get("text")
        if item.get("audio_wav") is not None:
            audio_bytes = item.get("audio_wav").tobytes()
            audio += audio_bytes
            with wave.open(
                f"{ASSETS_DIR}/qwen3omni_image_audio_interaction_stream_{i}.wav", "wb"
            ) as f:
                f.setnchannels(1)
                f.setsampwidth(2)
                f.setframerate(24000)
                f.writeframes(audio_bytes)

    print(text)
    with wave.open(f"{ASSETS_DIR}/qwen3omni_image_audio_interaction_stream.wav", "wb") as f:
        f.setnchannels(1)
        f.setsampwidth(2)
        f.setframerate(24000)
        f.writeframes(audio)


def audio_function_call(**kwargs):
    audio_path = (
        "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen3-Omni/cookbook/functioncall_case.wav"
    )

    messages = [
        {
            "role": "system",
            "content": """
    You may call one or more functions to assist with the user query.

    You are provided with function signatures within <tools></tools> XML tags:
    <tools>
    {'type': 'function', 'function': {'name': 'web_search', 'description': 'Utilize the web search engine to retrieve relevant information based on multiple queries.', 'parameters': {'type': 'object', 'properties': {'queries': {'type': 'array', 'items': {'type': 'string', 'description': 'The search query.'}, 'description': 'The list of search queries.'}}, 'required': ['queries']}}}
    {'type': 'function', 'function': {'name': 'car_ac_control', 'description': "Control the vehicle's air conditioning system to turn it on/off and set the target temperature", 'parameters': {'type': 'object', 'properties': {'temperature': {'type': 'number', 'description': 'Target set temperature in Celsius degrees'}, 'ac_on': {'type': 'boolean', 'description': 'Air conditioning status (true=on, false=off)'}}, 'required': ['temperature', 'ac_on']}}}
    </tools>

    For each function call, return a json object with function name and arguments within <tool_call></tool_call> XML tags:
    <tool_call>
    {{"name": <function-name>, "arguments": <args-json-object>}}
    </tool_call>""",
        },
        {"role": "user", "content": [{"type": "audio", "audio": audio_path}]},
    ]

    response, audio = run_model(
        messages=messages,
        return_audio=False,
        use_audio_in_video=False,
    )
    print(response)
    if audio is not None:
        audio_bytes = audio.tobytes()
        with wave.open(f"{ASSETS_DIR}/qwen3omni_audio_function_call.wav", "wb") as f:
            f.setnchannels(1)
            f.setsampwidth(2)
            f.setframerate(24000)
            f.writeframes(audio_bytes)


def text_image_video_audio_interaction(**kwargs):
    messages = [
        {
            "role": "user",
            "content": [
                {"type": "image", "image": os.path.join(ASSETS_DIR, "03-Confusing-Pictures.jpg")},
                {"type": "video", "video": os.path.join(ASSETS_DIR, "music.mp4")},
                {"type": "audio", "audio": os.path.join(ASSETS_DIR, "1272-128104-0000.flac")},
                {"type": "text", "text": "Analyze this audio, image, and video together."},
            ],
        }
    ]

    response, audio = run_model(
        messages=messages,
        return_audio=True,
        use_audio_in_video=True,
    )
    print(response)
    if audio is not None:
        audio_bytes = audio.tobytes()
        with wave.open(f"{ASSETS_DIR}/qwen3omni_text_image_video_audio_interaction.wav", "wb") as f:
            f.setnchannels(1)
            f.setsampwidth(2)
            f.setframerate(24000)
            f.writeframes(audio_bytes)


def batch_requests():
    """need return_audio=False"""
    # Conversation with video only
    conversation1 = [
        {
            "role": "user",
            "content": [
                {"type": "video", "video": os.path.join(ASSETS_DIR, "draw1.mp4")},
            ],
        },
    ]

    # Conversation with audio only
    conversation2 = [
        {
            "role": "user",
            "content": [
                {"type": "audio", "audio": os.path.join(ASSETS_DIR, "1272-128104-0000.flac")},
            ],
        },
    ]

    # Conversation with pure text
    conversation3 = [
        {"role": "user", "content": [{"type": "text", "text": "who are you?"}]},
    ]

    # Conversation with mixed media
    conversation4 = [
        {
            "role": "user",
            "content": [
                {"type": "image", "image": os.path.join(ASSETS_DIR, "03-Confusing-Pictures.jpg")},
                {"type": "video", "video": os.path.join(ASSETS_DIR, "music.mp4")},
                {"type": "audio", "audio": os.path.join(ASSETS_DIR, "1272-128104-0000.flac")},
                {"type": "text", "text": "Analyze this audio, image, and video together."},
            ],
        },
    ]

    # Combine messages for batch processing
    conversations = [conversation1, conversation2, conversation3, conversation4]
    texts, _ = run_model(conversations, return_audio=False, use_audio_in_video=True)
    print(texts)


def achatbot_generate():
    import torchaudio
    import soundfile as sf
    from achatbot.core.llm.transformers.manual_vision_voice_qwen3 import (
        TransformersManualQwen3OmniLLM,
    )
    from achatbot.common.session import Session, SessionCtx
    from achatbot.core.llm import LLMEnvInit
    from achatbot.common.logger import Logger

    Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False, is_console=True)

    session = Session(**SessionCtx("test_client_id", 16000, 2).__dict__)
    args = LLMEnvInit.get_qwen3omni_transformers_args()
    args["lm_model_name_or_path"] = "/root/.achatbot/models/Qwen/Qwen3-Omni-30B-A3B-Instruct"
    args["init_chat_prompt"] = SYSTEM_PROMPT
    args["verbose"] = True
    args["speaker"] = "Ethan"
    args["lm_attn_impl"] = "flash_attention_2"
    args["warmup_steps"] = 0
    args["warmup_prompt"] = "你叫什么名字？"
    args["thinker_args"]["lm_gen_temperature"] = 0.0
    args["code2wav_args"]["chunk_size"] = 50
    args["code2wav_args"]["left_context_size"] = 25
    llm = TransformersManualQwen3OmniLLM(**args)

    print("----start generate stream----")

    session.ctx.state["prompt"] = [
        {"type": "image", "image": os.path.join(ASSETS_DIR, "03-Confusing-Pictures.jpg")},
        # {"type": "audio", "audio": ""},
        # {"type": "video", "video": ""},
        {"type": "text", "text": "请描述一下图片中的内容"},
    ]
    kwargs = {
        "use_audio_in_video": False,
        "thinker_top_k": 10,
        "thinker_top_p": 0.9,
        "thinker_temperature": 0.95,
        "thinker_repetition_penalty": 1.1,
        "thinker_min_new_tokens": 1,
        "thinker_max_tokens_per_step": 15,
        "thinker_stop_strings_per_step": [",", ".", "，", "。"],
        "thinker_max_new_tokens": 150,
        "thinker_eos_token_ids": [
            151643,
            151645,
        ],
        "thinker_pad_token_id": 151643,
    }
    chunk_stream = llm.generate(session, **kwargs)
    gen_text = ""
    gen_all_text = ""
    audios = []
    times = []
    start_time = time.perf_counter()
    for i, chunk in enumerate(chunk_stream):
        times.append(time.perf_counter() - start_time)
        print(chunk)
        text = chunk["text"] if "text" in chunk else ""
        if gen_text != text:
            gen_text = text
            gen_all_text += gen_text
        if "audio_wav" in chunk:
            wav = chunk["audio_wav"]
            print(text, wav.shape)
            audios.append(wav.squeeze().cpu().numpy())
            # save_audio_path = os.path.join(ASSETS_DIR, f"achatbot_generate_stream-{i}-{text}.wav")
            # torchaudio.save(save_audio_path, wav, sample_rate=24000)
            # print(f"Audio saved to {save_audio_path}")
        else:
            print(text)
        start_time = time.perf_counter()

    print(f"gen all text: {gen_all_text}")
    if len(audios) > 0:
        save_audio_path = os.path.join(ASSETS_DIR, f"achatbot_generate_stream.wav")
        sf.write(save_audio_path, np.concatenate(audios), samplerate=24000)
        print(f"All Audio saved to {save_audio_path}")
        info = sf.info(save_audio_path, verbose=True)
        print(
            f"thinker->talker->code2wav chunk streaming first chunk time: {times[0]} s | wav duration: {info.duration} s | cost: {sum(times)} s | RTF: {sum(times) / info.duration}"
        )


"""
modal run src/llm/transformers/qwen3_omni.py --task tokenizer
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task dump_model
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task asr
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task text2speech
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task speech_translation
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task image_question
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task audio_interaction
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task audio_interaction_scene (speech chat)
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task video_interaction
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task video_interaction_scene (video include audio chat)
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task video_information_extracting
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task image_text_interaction
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task image_audio_interaction
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task audio_function_call
IMAGE_GPU=B200 modal run src/llm/transformers/qwen3_omni.py --task text_image_video_audio_interaction
IMAGE_GPU=B200 modal run src/llm/transformers/qwen3_omni.py --task batch_requests

# think chunk stream
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task text2text_stream
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task text2speech_stream
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task image_text_interaction_stream
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task image_audio_interaction_stream
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task audio_interaction_scene_stream (speech chat)
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task video_interaction_scene_stream (video include audio chat)

IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task code2wav
IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task code2wav_stream

ACHATBOT_VERSION=0.0.26.post1 IMAGE_GPU=A100-80GB modal run src/llm/transformers/qwen3_omni.py --task achatbot_generate

> [!TIP]:
> - 生成音频中未对特殊字符进行处理（omni统一到一起直接生成音频的弊端, 也许可以在隐藏层解决, 系统提示词限制貌似不起作用(提示不含特殊字符）, 比如：
    *   **优点**：这款饮料是专门为运动后设计的。它的核心成分是电解质，标签上明确写着“电解质≥200mg”，这能有效补充运动时因大量出汗流失的钠、钾等矿物质，帮助维持体液平衡，防止抽筋。同时，它也含有维生素E和维生素B6，有助于能量代谢。它还是0糖0卡的，不用担心额外的热量。

"""


@app.local_entrypoint()
def main(task: str = "tokenizer"):
    tasks = {
        "tokenizer": tokenizer,
        "dump_model": dump_model,
        "code2wav": code2wav,
        "code2wav_stream": code2wav_stream,
        "asr": asr,
        "text2text": text2text,
        "text2speech": text2speech,
        "speech_translation": speech_translation,
        "image_question": image_question,
        "audio_interaction": audio_interaction,
        "audio_interaction_scene": audio_interaction_scene,
        "video_interaction": video_interaction,
        "video_interaction_scene": video_interaction_scene,
        "video_information_extracting": video_information_extracting,
        "image_text_interaction": image_text_interaction,
        "image_audio_interaction": image_audio_interaction,
        "audio_function_call": audio_function_call,
        "text_image_video_audio_interaction": text_image_video_audio_interaction,
        "batch_requests": batch_requests,
        "text2text_stream": text2text_stream,
        "text2speech_stream": text2speech_stream,
        "image_text_interaction_stream": image_text_interaction_stream,
        "image_audio_interaction_stream": image_audio_interaction_stream,
        "audio_interaction_scene_stream": audio_interaction_scene_stream,
        "video_interaction_scene_stream": video_interaction_scene_stream,
        "achatbot_generate": achatbot_generate,
    }
    if task not in tasks:
        raise ValueError(f"task {task} not found")
    print(f"running task {task}")
    run.remote(tasks[task])



import logging
from threading import Thread
from time import perf_counter
import math
import re

from PIL import Image

try:
    import torch
    import torchvision.transforms as T
    from torchvision.transforms.functional import InterpolationMode
    from transformers import AutoTokenizer, AutoModel, AutoConfig
    from transformers.generation.streamers import TextIteratorStreamer

except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error(
        "In order to use Keye VLM, you need to `pip install achatbot[llm_transformers_manual_vision_skyworkr1v]`"
    )
    raise Exception(f"Missing module: {e}")


from src.common.utils.helper import get_device, print_model_params
from src.common.random import set_all_random_seed
from src.common.chat_history import ChatHistory
from src.common.session import Session
from src.types.speech.language import TO_LLM_LANGUAGE
from src.types.llm.transformers import TransformersLMArgs
from .base import TransformersBaseLLM


def split_model(model_path):
    device_map = {}
    world_size = torch.cuda.device_count()
    config = AutoConfig.from_pretrained(model_path, trust_remote_code=True)
    num_layers = config.llm_config.num_hidden_layers
    num_layers_per_gpu = math.ceil(num_layers / (world_size - 0.5))
    num_layers_per_gpu = [num_layers_per_gpu] * world_size
    num_layers_per_gpu[0] = math.ceil(num_layers_per_gpu[0] * 0.5)
    layer_cnt = 0
    for i, num_layer in enumerate(num_layers_per_gpu):
        for j in range(num_layer):
            device_map[f"language_model.model.layers.{layer_cnt}"] = i
            layer_cnt += 1
    device_map["vision_model"] = 0
    device_map["mlp1"] = 0
    device_map["language_model.model.tok_embeddings"] = 0
    device_map["language_model.model.embed_tokens"] = 0
    device_map["language_model.output"] = 0
    device_map["language_model.model.norm"] = 0
    device_map["language_model.model.rotary_emb"] = 0
    device_map["language_model.lm_head"] = 0
    device_map[f"language_model.model.layers.{num_layers - 1}"] = 0
    return device_map


IMAGENET_MEAN = (0.485, 0.456, 0.406)
IMAGENET_STD = (0.229, 0.224, 0.225)


def build_transform(input_size):
    MEAN, STD = IMAGENET_MEAN, IMAGENET_STD
    transform = T.Compose(
        [
            T.Lambda(lambda img: img.convert("RGB") if img.mode != "RGB" else img),
            T.Resize((input_size, input_size), interpolation=InterpolationMode.BICUBIC),
            T.ToTensor(),
            T.Normalize(mean=MEAN, std=STD),
        ]
    )
    return transform


def find_closest_aspect_ratio(aspect_ratio, target_ratios, width, height, image_size):
    best_ratio_diff = float("inf")
    best_ratio = (1, 1)
    area = width * height
    for ratio in target_ratios:
        target_aspect_ratio = ratio[0] / ratio[1]
        ratio_diff = abs(aspect_ratio - target_aspect_ratio)
        if ratio_diff < best_ratio_diff:
            best_ratio_diff = ratio_diff
            best_ratio = ratio
        elif ratio_diff == best_ratio_diff:
            if area > 0.5 * image_size * image_size * ratio[0] * ratio[1]:
                best_ratio = ratio
    return best_ratio


def dynamic_preprocess(image, min_num=1, max_num=12, image_size=448, use_thumbnail=False):
    orig_width, orig_height = image.size
    aspect_ratio = orig_width / orig_height
    target_ratios = set(
        (i, j)
        for n in range(min_num, max_num + 1)
        for i in range(1, n + 1)
        for j in range(1, n + 1)
        if i * j <= max_num and i * j >= min_num
    )
    target_ratios = sorted(target_ratios, key=lambda x: x[0] * x[1])
    target_aspect_ratio = find_closest_aspect_ratio(
        aspect_ratio, target_ratios, orig_width, orig_height, image_size
    )
    target_width = image_size * target_aspect_ratio[0]
    target_height = image_size * target_aspect_ratio[1]
    blocks = target_aspect_ratio[0] * target_aspect_ratio[1]
    resized_img = image.resize((target_width, target_height))
    processed_images = []
    for i in range(blocks):
        box = (
            (i % (target_width // image_size)) * image_size,
            (i // (target_width // image_size)) * image_size,
            ((i % (target_width // image_size)) + 1) * image_size,
            ((i // (target_width // image_size)) + 1) * image_size,
        )
        split_img = resized_img.crop(box)
        processed_images.append(split_img)
    assert len(processed_images) == blocks
    if use_thumbnail and len(processed_images) != 1:
        thumbnail_img = image.resize((image_size, image_size))
        processed_images.append(thumbnail_img)
    return processed_images


def load_image(image: Image.Image, input_size=448, max_num=12):
    image = image.convert("RGB")
    transform = build_transform(input_size=input_size)
    images = dynamic_preprocess(image, image_size=input_size, use_thumbnail=True, max_num=max_num)
    pixel_values = [transform(image) for image in images]
    pixel_values = torch.stack(pixel_values)
    return pixel_values


def get_prompt(conv_template, messages, thinking=True):
    assert messages[-1]["role"] == "user"

    for i, message in enumerate(messages):
        assert message.get("content")
        assert len(message.get("content")) > 0

        if message["role"] == "system":
            if message["content"][0]["text"]:
                conv_template.system_message = message["content"][0]["text"]
        elif message["role"] == "user":
            query = ""
            text = ""
            for item in message["content"]:
                if item["type"] == "text":
                    text = item["text"]
                if item["type"] == "image":
                    query += "<image>\n"
            query += text
            if i == len(messages) - 1:
                query = query.replace("<image>", "<IMAGE>")
            conv_template.append_message("user", query)
        elif message["role"] == "assistant":
            answer = ""
            for item in message["content"]:
                if item["type"] == "text":
                    answer += item["text"]
            conv_template.append_message("assistant", answer)
    conv_template.append_message("assistant", None)

    prompt = conv_template.get_prompt()
    if not prompt.endswith("\n<think>"):
        prompt += "\n<think>"
    if thinking is False:
        prompt = re.sub(r"\n<think>", "", prompt, count=1)

    return prompt


class TransformersManualVisionSkyworkR1V(TransformersBaseLLM):
    TAG = "llm_transformers_manual_vision_skyworkr1v"

    def __init__(self, **args) -> None:
        self.args = TransformersLMArgs(**args)
        gpu_prop = torch.cuda.get_device_properties("cuda")

        if self.args.lm_device_map is not None:
            if isinstance(self.args.lm_device_map, dict):
                customer_deivce_map = self.args.lm_device_map
                default_device_map = split_model(self.args.lm_model_name_or_path)
                self.args.lm_device_map = {**default_device_map, **customer_deivce_map}
            logging.info(f"TransformersLMArgs: {self.args}")
        if self.args.lm_device_map:
            self._model = AutoModel.from_pretrained(
                self.args.lm_model_name_or_path,
                torch_dtype=torch.bfloat16,
                #!NOTE: https://github.com/huggingface/transformers/issues/20896
                # device_map for multi cpu/gpu with accelerate
                device_map=self.args.lm_device_map,
                trust_remote_code=True,
                use_flash_attn=True if gpu_prop.major >= 8 else False,
                load_in_8bit=False,
                low_cpu_mem_usage=True,
            ).eval()
        else:
            self._model = (
                AutoModel.from_pretrained(
                    self.args.lm_model_name_or_path,
                    torch_dtype=torch.bfloat16,
                    trust_remote_code=True,
                    use_flash_attn=True if gpu_prop.major >= 8 else False,
                    load_in_8bit=False,
                    low_cpu_mem_usage=True,
                )
                .eval()
                .to(self.args.lm_device)
            )

        logging.info(f"TransformersLMArgs: {self.args}")
        print_model_params(self._model, self.TAG)
        self._tokenizer = AutoTokenizer.from_pretrained(
            self.args.lm_model_name_or_path,
            use_fast=True,
            trust_remote_code=True,
        )
        self._model.img_context_token_id = self._tokenizer.convert_tokens_to_ids("<IMG_CONTEXT>")
        self.eos_token_id = self._tokenizer.convert_tokens_to_ids(
            self._model.conv_template.sep.strip()
        )

        self._chat_history = ChatHistory(self.args.chat_history_size)  # no history
        if self.args.init_chat_role and self.args.init_chat_prompt:
            self._chat_history.init(
                {
                    "role": self.args.init_chat_role,
                    "content": [{"type": "text", "text": self.args.init_chat_prompt}],
                }
            )
        self.session_chat_history = {}

        self.warmup()

    def warmup(self):
        if self.args.warmup_steps <= 0:
            return
        dummy_input_text = self.args.warmup_prompt
        dummy_pil_image = Image.new("RGB", (100, 100), color="white")
        dummy_msgs = [
            {
                "role": self.args.user_role,
                "content": [
                    {"type": "text", "text": dummy_input_text},
                    {"type": "image", "image": dummy_pil_image},
                ],
            }
        ]

        tpl = self._model.conv_template.copy()
        prompt = get_prompt(tpl, dummy_msgs)

        pixel_values = []
        for item in dummy_msgs[-1]["content"]:
            if item["type"] == "image":
                image_file = item["image"]
                pixel_values.append(load_image(image_file))
        num_patches_list = [img.size(0) for img in pixel_values]
        pixel_values = (
            torch.cat(pixel_values, dim=0).to(torch.bfloat16).to(self._model.device)
            if len(pixel_values) > 0
            else None
        )
        pixel_values is not None and logging.info(f"{pixel_values.shape=}")

        # replace <image> placeholder
        # NOTE: just process curr user image query
        for num_patches in num_patches_list:
            image_tokens = (
                "<img>" + "<IMG_CONTEXT>" * self._model.num_image_token * num_patches + "</img>"
            )
            prompt = prompt.replace("<IMAGE>", image_tokens, 1)

        inputs = self._tokenizer([prompt], padding=True, return_tensors="pt").to(self._model.device)
        for key, value in inputs.items():
            logging.info(f"{key}: {value.shape=}") if isinstance(
                value, torch.Tensor
            ) else logging.info(f"{key}: {value} {value.dtype=}")

        streamer = TextIteratorStreamer(self._tokenizer, skip_prompt=True, skip_special_tokens=True)

        warmup_gen_kwargs = dict(
            **inputs,
            pixel_values=pixel_values,
            streamer=streamer,
            do_sample=True if self.args.lm_gen_temperature > 0 else False,
            temperature=self.args.lm_gen_temperature,
            top_k=self.args.lm_gen_top_k,
            top_p=self.args.lm_gen_top_p,
            repetition_penalty=self.args.lm_gen_repetition_penalty,
            min_new_tokens=1,
            max_new_tokens=128,
            eos_token_id=self.eos_token_id,
        )

        self._warmup(
            target=self._model.generate,
            kwargs=warmup_gen_kwargs,
            streamer=streamer,
        )

    @torch.inference_mode()
    def generate(self, session: Session, **kwargs):
        enable_thinking = kwargs.get("thinking", self.args.lm_gen_thinking)
        if enable_thinking is None:  # default thinking is True
            enable_thinking = True
        seed = kwargs.get("seed", self.args.lm_gen_seed)
        set_all_random_seed(seed)

        prompt = session.ctx.state["prompt"]
        assert len(prompt) > 0
        message = {"role": self.args.user_role, "content": prompt}
        self.add_chat_history(session, message)
        chat_history = self.get_session_chat_history(session.ctx.client_id)
        logging.info(f"{session.ctx.client_id} chat_history:{chat_history}")

        tpl = self._model.conv_template.copy()
        prompt = get_prompt(tpl, chat_history, thinking=enable_thinking)

        pixel_values = []
        for item in chat_history[-1]["content"]:
            if item["type"] == "image":
                image_file = item["image"]
                pixel_values.append(load_image(image_file))
        num_patches_list = [img.size(0) for img in pixel_values]
        pixel_values = (
            torch.cat(pixel_values, dim=0).to(torch.bfloat16).to(self._model.device)
            if len(pixel_values) > 0
            else None
        )
        pixel_values is not None and logging.info(f"{pixel_values.shape=}")
        logging.info(f"{prompt=}")

        # replace <image> placeholder
        # NOTE: just process current user image query
        for num_patches in num_patches_list:
            image_tokens = (
                "<img>" + "<IMG_CONTEXT>" * self._model.num_image_token * num_patches + "</img>"
            )
            prompt = prompt.replace("<IMAGE>", image_tokens, 1)

        inputs = self._tokenizer([prompt], padding=True, return_tensors="pt").to(self._model.device)
        for key, value in inputs.items():
            logging.info(f"{key}: {value.shape=}") if isinstance(
                value, torch.Tensor
            ) else logging.info(f"{key}: {value} {value.dtype=}")

        streamer = TextIteratorStreamer(self._tokenizer, skip_prompt=True, skip_special_tokens=True)

        generation_kwargs = dict(
            **inputs,
            pixel_values=pixel_values,
            streamer=streamer,
            do_sample=True
            if kwargs.get("temperature", self.args.lm_gen_temperature) > 0
            else False,
            temperature=kwargs.get("temperature", self.args.lm_gen_temperature),
            top_k=kwargs.get("top_k", self.args.lm_gen_top_k),
            top_p=kwargs.get("top_p", self.args.lm_gen_top_p),
            repetition_penalty=kwargs.get(
                "repetition_penalty", self.args.lm_gen_repetition_penalty
            ),
            min_new_tokens=kwargs.get("min_new_tokens", self.args.lm_gen_min_new_tokens),
            max_new_tokens=kwargs.get("max_new_tokens", self.args.lm_gen_max_new_tokens),
            eos_token_id=self.eos_token_id,
        )
        thread = Thread(target=self._model.generate, kwargs=generation_kwargs)
        thread.start()

        generated_text = ""
        start = perf_counter()
        times = []
        is_output_think = self.args.lm_gen_think_output
        is_thinking = False
        is_answer = True
        think_text = ""
        for new_text in streamer:
            times.append(perf_counter() - start)
            if (
                ("<think>" in new_text or enable_thinking is True)
                and is_thinking is False
                and think_text == ""
            ):
                yield "思考中，请稍等。"
                is_thinking = True
            if "</think>" in new_text:
                is_thinking = False
                think_text += new_text
                logging.info(f"{think_text=}")
                new_text = new_text.replace("</think>", "")
                is_answer = True
            if is_thinking is True:
                think_text += new_text
                if is_output_think is True:
                    generated_text += new_text
                    yield new_text
                else:
                    yield None
                start = perf_counter()
                continue

            if "<|im_end|>" in new_text:
                is_answer = False
                start = perf_counter()
                break

            if is_answer is True:
                generated_text += new_text
                yield new_text
            start = perf_counter()
        yield "."  # end the sentence for downstream process sentence, e.g.: tts
        logging.info(f"{generated_text=} TTFT: {times[0]:.4f}s total time: {sum(times):.4f}s")
        torch.cuda.empty_cache()
        self.session_chat_history[session.ctx.client_id].append(
            {"role": "assistant", "content": [{"type": "text", "text": generated_text}]}
        )



import urllib.parse
import os
import time
import urllib
import requests

from dotenv import load_dotenv

load_dotenv()


daily_api_path = os.getenv("DAILY_API_URL") or "api.daily.co/v1"
daily_api_key = os.getenv("DAILY_API_KEY")


def create_room(name: str | None = None) -> tuple[str, str]:
    """
    Helper function to create a Daily room.
    # See: https://docs.daily.co/reference/rest-api/rooms

    Returns:
        tuple: A tuple containing the room URL and room name.

    Raises:
        Exception: If the request to create the room fails or if the response does not contain the room URL or room name.
    """
    room_props = {
        "exp": time.time() + 60 * 60,  # 1 hour
        "enable_chat": True,
        "enable_emoji_reactions": True,
        "eject_at_room_exp": True,
        "enable_prejoin_ui": False,  # Important for the bot to be able to join headlessly
    }
    json_data = {"properties": room_props}
    if name is not None:
        json_data["name"] = name
    res = requests.post(
        f"https://{daily_api_path}/rooms",
        headers={"Authorization": f"Bearer {daily_api_key}"},
        json=json_data,
    )
    if res.status_code != 200:
        raise Exception(f"Unable to create room: {res.text}")

    data = res.json()
    room_url: str = data.get("url")
    room_name: str = data.get("name")
    if room_url is None or room_name is None:
        raise Exception("Missing room URL or room name in response")

    return room_url, room_name


def get_name_from_url(room_url: str) -> str:
    """
    Extracts the name from a given room URL.

    Args:
        room_url (str): The URL of the room.

    Returns:
        str: The extracted name from the room URL.
    """
    return urllib.parse.urlparse(room_url).path[1:]


def get_token(room_url: str) -> str:
    """
    Retrieves a meeting token for the specified Daily room URL.
    # See: https://docs.daily.co/reference/rest-api/meeting-tokens

    Args:
        room_url (str): The URL of the Daily room.

    Returns:
        str: The meeting token.

    Raises:
        Exception: If no room URL is specified or if no Daily API key is specified.
        Exception: If there is an error creating the meeting token.
    """
    if not room_url:
        raise Exception(
            "No Daily room specified. You must specify a Daily room in order a token to be generated."
        )

    if not daily_api_key:
        raise Exception(
            "No Daily API key specified. set DAILY_API_KEY in your environment to specify a Daily API key, available from https://dashboard.daily.co/developers."
        )

    expiration: float = time.time() + 60 * 60
    room_name = get_name_from_url(room_url)

    res: requests.Response = requests.post(
        f"https://{daily_api_path}/meeting-tokens",
        headers={"Authorization": f"Bearer {daily_api_key}"},
        json={
            "properties": {
                "room_name": room_name,
                "is_owner": True,  # Owner tokens required for transcription
                "exp": expiration,
            }
        },
    )

    if res.status_code != 200:
        raise Exception(f"Failed to create meeting token: {res.status_code} {res.text}")

    token: str = res.json()["token"]

    return token
